# - coding: utf-8 -

# Copyright (C) 2008-2010 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject as gobject
from gi.repository import Gtk as gtk
from gi.repository import Pango as pango

import time
from hamster.lib import graphics, stuff
from hamster.lib import datetime as dt
import locale

class Bar(graphics.Sprite):
    def __init__(self, key, value, normalized, label_color):
        graphics.Sprite.__init__(self, cache_as_bitmap=True)
        self.key, self.value, self.normalized = key, value, normalized

        self.height = 0
        self.width = 20
        self.interactive = True
        self.fill = None

        self.label = graphics.Label(value, size=8, color=label_color)
        self.label_background = graphics.Rectangle(self.label.width + 4, self.label.height + 4, 4, visible=False)
        self.add_child(self.label_background)
        self.add_child(self.label)
        self.connect("on-render", self.on_render)

    def on_render(self, sprite):
        # invisible rectangle for the mouse, covering whole area
        self.graphics.rectangle(0, 0, self.width, self.height)
        self.graphics.fill("#000", 0)

        size = round(self.width * self.normalized)

        self.graphics.rectangle(0, 0, size, self.height, 3)
        self.graphics.rectangle(0, 0, min(size, 3), self.height)
        self.graphics.fill(self.fill)

        self.label.y = (self.height - self.label.height) / 2

        horiz_offset = min(10, self.label.y * 2)

        if self.label.width < size - horiz_offset * 2:
            #if it fits in the bar
            self.label.x = size - self.label.width - horiz_offset
        else:
            self.label.x = size + 3

        self.label_background.x = self.label.x - 2
        self.label_background.y = self.label.y - 2


class Chart(graphics.Scene):
    __gsignals__ = {
        "bar-clicked": (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, )),
    }

    def __init__(self, max_bar_width = 20, legend_width = 70, value_format = "%.2f", interactive = True):
        graphics.Scene.__init__(self)

        self.selected_keys = [] # keys of selected bars

        self.bars = []
        self.labels = []
        self.data = None

        self.max_width = max_bar_width
        self.legend_width = legend_width
        self.value_format = value_format
        self.graph_interactive = interactive

        self.plot_area = graphics.Sprite(interactive = False)
        self.add_child(self.plot_area)

        self.bar_color, self.label_color = None, None

        self.connect("on-enter-frame", self.on_enter_frame)

        if self.graph_interactive:
            self.connect("on-mouse-over", self.on_mouse_over)
            self.connect("on-mouse-out", self.on_mouse_out)
            self.connect("on-click", self.on_click)

    def find_colors(self):
        bg_color = "#eee" #self.get_style().bg[gtk.StateType.NORMAL].to_string()
        self.bar_color = self.colors.contrast(bg_color, 30)

        # now for the text - we want reduced contrast for relaxed visuals
        fg_color = "#aaa" #self.get_style().fg[gtk.StateType.NORMAL].to_string()
        self.label_color = self.colors.contrast(fg_color,  80)


    def on_mouse_over(self, scene, bar):
        if bar.key not in self.selected_keys:
            bar.fill = "#999" #self.get_style().base[gtk.StateType.PRELIGHT].to_string()

    def on_mouse_out(self, scene, bar):
        if bar.key not in self.selected_keys:
            bar.fill = self.bar_color

    def on_click(self, scene, event, clicked_bar):
        if not clicked_bar: return
        self.emit("bar-clicked", clicked_bar.key)

    def plot(self, keys, data):
        self.data = data

        bars = dict([(bar.key, bar.normalized) for bar in self.bars])

        max_val = float(max(data or [0]))

        new_bars, new_labels = [], []
        for key, value in zip(keys, data):
            if max_val:
                normalized = value / max_val
            else:
                normalized = 0
            bar = Bar(key, locale.format(self.value_format, value), normalized, self.label_color)
            bar.interactive = self.graph_interactive

            if key in bars:
                bar.normalized = bars[key]
                self.tweener.add_tween(bar, normalized=normalized)
            new_bars.append(bar)

            label = graphics.Label(stuff.escape_pango(key), size = 8, alignment = pango.Alignment.RIGHT)
            new_labels.append(label)


        self.plot_area.remove_child(*self.bars)
        self.remove_child(*self.labels)

        self.bars, self.labels = new_bars, new_labels
        self.add_child(*self.labels)
        self.plot_area.add_child(*self.bars)

        self.show()
        self.redraw()


    def on_enter_frame(self, scene, context):
        # adjust sizes and positions on redraw

        legend_width = self.legend_width
        if legend_width < 1: # allow fractions
            legend_width = int(self.width * legend_width)

        self.find_colors()

        self.plot_area.y = 0
        self.plot_area.height = self.height - self.plot_area.y
        self.plot_area.x = legend_width + 8
        self.plot_area.width = self.width - self.plot_area.x

        y = 0
        for i, (label, bar) in enumerate(zip(self.labels, self.bars)):
            bar_width = min(round((self.plot_area.height - y) / (len(self.bars) - i)), self.max_width)
            bar.y = y
            bar.height = bar_width
            bar.width = self.plot_area.width

            if bar.key in self.selected_keys:
                bar.fill = "#aaa" #self.get_style().bg[gtk.StateType.SELECTED].to_string()

                if bar.normalized == 0:
                    bar.label.color = "#666" #self.get_style().fg[gtk.StateType.SELECTED].to_string()
                    bar.label_background.fill = "#aaa" #self.get_style().bg[gtk.StateType.SELECTED].to_string()
                    bar.label_background.visible = True
                else:
                    bar.label_background.visible = False
                    if bar.label.x < round(bar.width * bar.normalized):
                        bar.label.color = "#666" #self.get_style().fg[gtk.StateType.SELECTED].to_string()
                    else:
                        bar.label.color = self.label_color

            if not bar.fill:
                bar.fill = self.bar_color

                bar.label.color = self.label_color
                bar.label_background.fill = None

            label.y = y + (bar_width - label.height) / 2 + self.plot_area.y

            label.width = legend_width
            if not label.color:
                label.color = self.label_color

            y += bar_width + 1




class HorizontalDayChart(graphics.Scene):
    """Pretty much a horizontal bar chart, except for values it expects tuple
    of start and end time, and the whole thing hangs in air"""
    def __init__(self, max_bar_width, legend_width):
        graphics.Scene.__init__(self)
        self.max_bar_width = max_bar_width
        self.legend_width = legend_width
        self.start_time, self.end_time = None, None
        self.connect("on-enter-frame", self.on_enter_frame)

    def plot_day(self, keys, data, start_time = None, end_time = None):
        self.keys, self.data = keys, data
        self.start_time, self.end_time = start_time, end_time
        self.show()
        self.redraw()

    def on_enter_frame(self, scene, context):
        g = graphics.Graphics(context)

        rowcount, keys = len(self.keys), self.keys

        start_hour = 0
        if self.start_time:
            start_hour = self.start_time
        end_hour = 24 * 60
        if self.end_time:
            end_hour = self.end_time


        # push graph to the right, so it doesn't overlap
        legend_width = self.legend_width or self.longest_label(keys)

        self.graph_x = legend_width
        self.graph_x += 8 #add another 8 pixes of padding

        self.graph_width = self.width - self.graph_x

        # TODO - should handle the layout business in graphics
        self.layout = context.create_layout()
        default_font = "Sans Serif" #pango.FontDescription(self.get_style().font_desc.to_string())
        default_font.set_size(8 * pango.SCALE)
        self.layout.set_font_description(default_font)


        #on the botttom leave some space for label
        self.layout.set_text("1234567890:")
        label_w, label_h = self.layout.get_pixel_size()

        self.graph_y, self.graph_height = 0, self.height - label_h - 4

        if not self.data:  #if we have nothing, let's go home
            return


        positions = {}
        y = 0
        bar_width = min(self.graph_height / float(len(self.keys)), self.max_bar_width)
        for i, key in enumerate(self.keys):
            positions[key] = (y + self.graph_y, round(bar_width - 1))

            y = y + round(bar_width)
            bar_width = min(self.max_bar_width,
                            (self.graph_height - y) / float(max(1, len(self.keys) - i - 1)))



        max_bar_size = self.graph_width - 15


        # now for the text - we want reduced contrast for relaxed visuals
        fg_color = "#666" #self.get_style().fg[gtk.StateType.NORMAL].to_string()
        label_color = self.colors.contrast(fg_color,  80)

        self.layout.set_alignment(pango.Alignment.RIGHT)
        self.layout.set_ellipsize(pango.ELLIPSIZE_END)

        # bars and labels
        self.layout.set_width(legend_width * pango.SCALE)

        factor = max_bar_size / float(end_hour - start_hour)

        # determine bar color
        bg_color = "#eee" #self.get_style().bg[gtk.StateType.NORMAL].to_string()
        base_color = self.colors.contrast(bg_color,  30)

        for i, label in enumerate(keys):
            g.set_color(label_color)

            self.layout.set_text(label)
            label_w, label_h = self.layout.get_pixel_size()

            context.move_to(0, positions[label][0] + (positions[label][1] - label_h) / 2)
            context.show_layout(self.layout)

            if isinstance(self.data[i], list) == False:
                self.data[i] = [self.data[i]]

            for row in self.data[i]:
                bar_x = round((row[0]- start_hour) * factor)
                bar_size = round((row[1] - start_hour) * factor - bar_x)

                g.fill_area(round(self.graph_x + bar_x),
                              positions[label][0],
                              bar_size,
                              positions[label][1],
                              base_color)

        #white grid and scale values
        self.layout.set_width(-1)

        context.set_line_width(1)

        pace = ((end_hour - start_hour) / 3) / 60 * 60
        last_position = positions[keys[-1]]


        grid_color = "#aaa" # self.get_style().bg[gtk.StateType.NORMAL].to_string()

        for i in range(start_hour + 60, end_hour, pace):
            x = round((i - start_hour) * factor)

            minutes = i % (24 * 60)

            self.layout.set_markup(dt.time(minutes / 60, minutes % 60).strftime("%H<small><sup>%M</sup></small>"))
            label_w, label_h = self.layout.get_pixel_size()

            context.move_to(self.graph_x + x - label_w / 2,
                            last_position[0] + last_position[1] + 4)
            g.set_color(label_color)
            context.show_layout(self.layout)


            g.set_color(grid_color)
            g.move_to(round(self.graph_x + x) + 0.5, self.graph_y)
            g.line_to(round(self.graph_x + x) + 0.5,
                                 last_position[0] + last_position[1])


        context.stroke()



# - coding: utf-8 -

# Copyright (C) 2008-2009 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)   # noqa: E402

import bisect
import cairo
import re

from gi.repository import Gdk as gdk
from gi.repository import Gtk as gtk
from gi.repository import GObject as gobject
from gi.repository import PangoCairo as pangocairo
from gi.repository import Pango as pango
from collections import defaultdict
from copy import deepcopy

from hamster import client
from hamster.lib import datetime as dt
from hamster.lib import stuff
from hamster.lib import graphics
from hamster.lib.configuration import runtime
from hamster.lib.fact import Fact


# note: Still experimenting in this module.
#       Code redundancy to be removed later.


def extract_search(text):
    fact = Fact.parse(text)
    search = fact.activity
    if fact.category:
        search += "@%s" % fact.category
    if fact.tags:
        search += " #%s" % (" #".join(fact.tags))
    return search

class DataRow(object):
    """want to split out visible label, description, activity data
      and activity data with time (full_data)"""
    def __init__(self, label, data=None, full_data=None, description=None):
        self.label = label
        self.data = data or label
        self.full_data = full_data or data or label
        self.description = description or ""

class Label(object):
    """a much cheaper label that would be suitable for cellrenderer"""
    def __init__(self, x=0, y=0):
        self.x, self.y = x, y
        self._label_context = cairo.Context(cairo.ImageSurface(cairo.FORMAT_A1, 0, 0))
        self.layout = pangocairo.create_layout(self._label_context)
        self.layout.set_font_description(pango.FontDescription(graphics._font_desc))
        self.layout.set_markup("Hamster") # dummy
        self.height = self.layout.get_pixel_size()[1]

    def show(self, g, text, color=None):
        g.move_to(self.x, self.y)

        self.layout.set_markup(text)
        g.save_context()
        if color:
            g.set_color(color)
        pangocairo.show_layout(g.context, self.layout)
        g.restore_context()


class CompleteTree(graphics.Scene):
    """
    ASCII Art

    | Icon | Activity - description |

    """

    __gsignals__ = {
        # enter or double-click, passes in current day and fact
        'on-select-row': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),
    }


    def __init__(self):
        graphics.Scene.__init__(self, style_class=gtk.STYLE_CLASS_VIEW)

        self.set_can_focus(False)

        self.row_positions = []

        self.current_row = None
        self.rows = []

        self.style = self._style

        self.label = Label(x=5, y=3)
        self.row_height = self.label.height + 10

        self.connect("on-key-press", self.on_key_press)
        self.connect("on-enter-frame", self.on_enter_frame)
        self.connect("on-mouse-move", self.on_mouse_move)
        self.connect("on-mouse-down", self.on_mouse_down)

    def _get_mouse_row(self, event):
        hover_row = None
        for row, y in zip(self.rows, self.row_positions):
            if y <= event.y <= (y + self.row_height):
                hover_row = row
                break
        return hover_row

    def on_mouse_move(self, scene, event):
        row = self._get_mouse_row(event)
        if row:
            self.current_row = row
            self.redraw()

    def on_mouse_down(self, scene, event):
        row = self._get_mouse_row(event)
        if row:
            self.set_current_row(self.rows.index(row))

    def on_key_press(self, scene, event):
        if event.keyval == gdk.KEY_Up:
            idx = self.rows.index(self.current_row) if self.current_row else 1
            self.set_current_row(idx - 1)

        elif event.keyval == gdk.KEY_Down:
            idx = self.rows.index(self.current_row) if self.current_row else -1
            self.set_current_row(idx + 1)

    def set_current_row(self, idx):
        idx = max(0, min(len(self.rows) - 1, idx))
        row = self.rows[idx]
        self.current_row = row
        self.redraw()
        self.emit("on-select-row", row)

    def set_rows(self, rows):
        self.current_row = None
        self.rows = rows
        self.set_row_positions()

    def set_row_positions(self):
        """creates a list of row positions for simpler manipulation"""
        self.row_positions = [i * self.row_height for i in range(len(self.rows))]
        self.set_size_request(0, self.row_positions[-1] + self.row_height if self.row_positions else 0)

    def on_enter_frame(self, scene, context):
        if not self.height:
            return

        colors = {
            "normal": self.style.get_color(gtk.StateFlags.NORMAL),
            "normal_bg": self.style.get_background_color(gtk.StateFlags.NORMAL),
            "selected": self.style.get_color(gtk.StateFlags.SELECTED),
            "selected_bg": self.style.get_background_color(gtk.StateFlags.SELECTED),
        }

        g = graphics.Graphics(context)
        g.set_line_style(1)
        g.translate(0.5, 0.5)


        for row, y in zip(self.rows, self.row_positions):
            g.save_context()
            g.translate(0, y)

            color, bg = colors["normal"], colors["normal_bg"]
            if row == self.current_row:
                color, bg = colors["selected"], colors["selected_bg"]
                g.fill_area(0, 0, self.width, self.row_height, bg)

            label = row.label
            if row.description:
                description_color = graphics.Colors.mix(bg, color, 0.75)
                description_color_str = graphics.Colors.hex(description_color)

                label = '{}  <span color="{}">[{}]</span>'.format(label,
                                                                  description_color_str,
                                                                  row.description)

            self.label.show(g, label, color=color)

            g.restore_context()



class CmdLineEntry(gtk.Entry):
    def __init__(self, *, parent, **kwargs):
        gtk.Entry.__init__(self, parent=parent, **kwargs)

        # default day for times without date
        self.default_day = None

        # to be set by the caller, if editing an existing fact
        self.original_fact = None

        self.popup = gtk.Window(type = gtk.WindowType.POPUP)
        self.popup.set_type_hint(gdk.WindowTypeHint.COMBO)  # why not
        self.popup.set_attached_to(self)  # attributes
        self.popup.set_transient_for(self.get_ancestor(gtk.Window))  # position

        box = gtk.Frame()
        box.set_shadow_type(gtk.ShadowType.IN)
        self.popup.add(box)

        self.complete_tree = CompleteTree()
        self.tree_checker = self.complete_tree.connect("on-select-row", self.on_tree_select_row)
        self.complete_tree.connect("on-click", self.on_tree_click)
        box.add(self.complete_tree)

        self.storage = client.Storage()
        self.load_suggestions()
        self.ignore_stroke = False

        self.set_icon_from_icon_name(gtk.EntryIconPosition.SECONDARY, "go-down-symbolic")

        self.checker = self.connect("changed", self.on_changed)
        self.connect("key-press-event", self.on_key_press)
        self.connect("focus-out-event", self.on_focus_out)
        self.connect("icon-press", self.on_icon_press)



    def on_changed(self, entry):
        text = self.get_text()

        with self.complete_tree.handler_block(self.tree_checker):
            self.show_suggestions(text)

        if self.ignore_stroke:
            self.ignore_stroke = False
            return

        def complete():
            text, suffix = self.complete_first()
            if suffix:
                #self.ignore_stroke = True
                with self.handler_block(self.checker):
                    self.update_entry("%s%s" % (text, suffix))
                    self.select_region(len(text), -1)
        gobject.timeout_add(0, complete)

    def on_focus_out(self, entry, event):
        self.popup.hide()

    def on_icon_press(self, entry, icon, event):
        if self.popup.get_visible():
            self.popup.hide()
        else:
            self.grab_focus()
            self.show_suggestions(self.get_text())

    def on_key_press(self, entry, event=None):
        if event.keyval in (gdk.KEY_BackSpace, gdk.KEY_Delete):
            self.ignore_stroke = True

        elif event.keyval in (gdk.KEY_Return, gdk.KEY_KP_Enter, gdk.KEY_Escape):
            self.popup.hide()
            self.set_position(-1)

        elif event.keyval in (gdk.KEY_Up, gdk.KEY_Down):
            if not self.popup.get_visible():
                self.show_suggestions(self.get_text())
            self.complete_tree.on_key_press(self, event)
            return True

    def on_tree_click(self, entry, tree, event):
        self.popup.hide()

    def on_tree_select_row(self, tree, row):
        with self.handler_block(self.checker):
            label = row.full_data
            self.update_entry(label)
            self.set_position(-1)


    def load_suggestions(self):
        self.todays_facts = self.storage.get_todays_facts()

        # list of facts of last month
        now = dt.datetime.now()
        last_month = self.storage.get_facts(now - dt.timedelta(days=30), now)

        # naive recency and frequency rank
        # score is as simple as you get 30-days_ago points for each occurence
        suggestions = defaultdict(int)
        for fact in last_month:
            days = 30 - (now - dt.datetime.combine(fact.date, dt.time())).total_seconds() / 60 / 60 / 24
            label = fact.activity
            if fact.category:
                label += "@%s" % fact.category

            suggestions[label] += days

            if fact.tags:
                label += " #%s" % (" #".join(fact.tags))
                suggestions[label] += days

        for rec in self.storage.get_activities():
            label = rec["name"]
            if rec["category"]:
                label += "@%s" % rec["category"]
            suggestions[label] += 0

        # list of (label, score), higher scores first
        self.suggestions = sorted(suggestions.items(), key=lambda x: x[1], reverse=True)

    def complete_first(self):
        text = self.get_text()
        fact = Fact.parse(text)
        search = extract_search(text)
        if not self.complete_tree.rows or not fact.activity:
            return text, None

        label = self.complete_tree.rows[0].data
        if label.startswith(search):
            return text, label[len(search):]

        return text, None


    def update_entry(self, text):
        self.set_text(text or "")


    def update_suggestions(self, text=""):
        """
            * from previous activity | set time | minutes ago | start now
            * to ongoing | set time

            * activity
            * [@category]
            * #tags, #tags, #tags

            * we will leave description for later

            all our magic is space separated, strictly, start-end can be just dash

            phases:

            [start_time] | [-end_time] | activity | [@category] | [#tag]
        """

        res = []

        fact = Fact.parse(text)
        now = dt.datetime.now()

        # figure out what we are looking for
        # time -> activity[@category] -> tags -> description
        # presence of an attribute means that we are not looking for the previous one
        # we still might be looking for the current one though
        looking_for = "start_time"
        fields = ["start_time", "end_time", "activity", "category", "tags", "description", "done"]
        for field in reversed(fields):
            if getattr(fact, field, None):
                looking_for = field
                if text[-1] == " ":
                    looking_for = fields[fields.index(field)+1]
                break


        fragments = [f for f in re.split("[\s|#]", text)]
        current_fragment = fragments[-1] if fragments else ""


        search = extract_search(text)

        matches = []
        for match, score in self.suggestions:
            if search in match:
                if match.startswith(search):
                    score += 10**8 # boost beginnings
                matches.append((match, score))

        # need to limit these guys, sorry
        matches = sorted(matches, key=lambda x: x[1], reverse=True)[:7]

        for match, score in matches:
            label = (fact.start_time or now).strftime("%H:%M")
            if fact.end_time:
                label += fact.end_time.strftime("-%H:%M")

            markup_label = label + " " + (stuff.escape_pango(match).replace(search, "<b>%s</b>" % search) if search else match)
            label += " " + match

            res.append(DataRow(markup_label, match, label))

        # list of tuples (description, variant)
        variants = []

        variant_fact = None
        if fact.end_time is None:
            description = "stop now"
            variant_fact = fact.copy()
            variant_fact.end_time = now
        elif self.todays_facts and fact == self.todays_facts[-1]:
            # that one is too dangerous, except for the last entry
            description = "keep up"
            # Do not use Fact(..., end_time=None): it would be a no-op
            variant_fact = fact.copy()
            variant_fact.end_time = None

        if variant_fact:
            variant_fact.description = None
            variant = variant_fact.serialized(default_day=self.default_day)
            variants.append((description, variant))

        if fact.start_time is None:
            description = "start now"
            variant = now.strftime("%H:%M ")
            variants.append((description, variant))

            prev_fact = self.todays_facts[-1] if self.todays_facts else None
            if prev_fact and prev_fact.end_time:
                since = (now - prev_fact.end_time).format()
                description = "from previous activity, %s ago" % since
                variant = prev_fact.end_time.strftime("%H:%M ")
                variants.append((description, variant))

            description = "start activity -n minutes ago (1 or 3 digits allowed)"
            variant = "-"
            variants.append((description, variant))

        text = text.strip()
        if text:
            description = "clear"
            variant = ""
            variants.append((description, variant))

        for (description, variant) in variants:
            res.append(DataRow(variant, description=description))

        self.complete_tree.set_rows(res)


    def show_suggestions(self, text):
        if not self.get_window():
            return

        entry_alloc = self.get_allocation()
        entry_x, entry_y = self.get_window().get_origin()[1:]
        x, y = entry_x + entry_alloc.x, entry_y + entry_alloc.y + entry_alloc.height

        self.popup.show_all()

        self.update_suggestions(text)

        tree_w, tree_h = self.complete_tree.get_size_request()

        self.popup.move(x, y)
        self.popup.resize(entry_alloc.width, tree_h)
        self.popup.show_all()


class ActivityEntry():
    """Activity entry widget.

    widget (gtk.Entry): the associated activity entry
    category_widget (gtk.Entry): the associated category entry
    """
    def __init__(self, widget=None, category_widget=None, **kwds):
        # widget and completion may be defined already
        # e.g. in the glade edit_activity.ui file
        self.widget = widget
        if not self.widget:
            self.widget = gtk.Entry(**kwds)

        self.category_widget = category_widget

        # internal list of actions added to the suggestions
        self._action_list = []
        self.completion = self.widget.get_completion()
        if not self.completion:
            self.completion = gtk.EntryCompletion()
            self.widget.set_completion(self.completion)

        # text to display/filter on, activity, category
        self.text_column = 0
        self.activity_column = 1
        self.category_column = 2

        # whether the category choice limit the activity suggestions
        self.filter_on_category = True if self.category_widget else False
        self.model = gtk.ListStore(str, str, str)
        self.completion.set_model(self.model)
        self.completion.set_text_column(self.text_column)
        self.completion.set_match_func(self.match_func, None)
        # enable selection with up and down arrow
        self.completion.set_inline_selection(True)
        # It is not possible to change actions later dynamically;
        # once actions are removed,
        # they can not be added back (they are not visible).
        # => nevermind, showing all actions.
        self.add_action("show all", "Show all activities")
        self.add_action("filter on category", "Filter on selected category")

        self.connect("icon-release", self.on_icon_release)
        self.connect("focus-in-event", self.on_focus_in_event)
        self.completion.connect('match-selected', self.on_match_selected)
        self.completion.connect("action_activated", self.on_action_activated)

    def add_action(self, name, text):
        """Add an action to the suggestions.

        name (str): unique label, use to retrieve the action index.
        text (str): text used to display the action.
        """
        markup = "<i>{}</i>".format(stuff.escape_pango(text))
        idx = len(self._action_list)
        self.completion.insert_action_markup(idx, markup)
        self._action_list.append(name)

    def clear(self, notify=True):
        self.widget.set_text("")
        if notify:
            self.emit("changed")

    def match_func(self, completion, key, iter, *user_data):
        if not key.strip():
            # show all keys if entry is empty
            return True
        else:
            # return whether the entered string is
            # anywhere in the first column data
            stripped_key = key.strip()
            activities = self.model.get_value(iter, self.activity_column).lower()
            categories = self.model.get_value(iter, self.category_column).lower()
            key_in_activity = stripped_key in activities
            key_in_category = stripped_key in categories
            return key_in_activity or key_in_category

    def on_action_activated(self, completion, index):
        name = self._action_list[index]
        if name == "clear":
            self.clear(notify=False)
        elif name == "show all":
            self.filter_on_category = False
            self.populate_completions()
        elif name == "filter on category":
            self.filter_on_category = True
            self.populate_completions()

    def on_focus_in_event(self, widget, event):
        self.populate_completions()

    def on_icon_release(self, entry, icon_pos, event):
        self.grab_focus()
        self.set_text("")
        self.emit("changed")

    def on_match_selected(self, entry, model, iter):
        activity_name = model[iter][self.activity_column]
        category_name = model[iter][self.category_column]
        combined = model[iter][self.text_column]
        if self.category_widget:
            self.set_text(activity_name)
            if not self.filter_on_category:
                self.category_widget.set_text(category_name)
        else:
            self.set_text(combined)
        return True  # prevent the standard callback from overwriting text

    def populate_completions(self):
        self.model.clear()
        if self.filter_on_category:
            category_names = [self.category_widget.get_text()]
        else:
            category_names = [category['name']
                              for category in runtime.storage.get_categories()]
        for category_name in category_names:
            category_id = runtime.storage.get_category_id(category_name)
            activities = runtime.storage.get_category_activities(category_id)
            for activity in activities:
                activity_name = activity["name"]
                text = "{}@{}".format(activity_name, category_name)
                self.model.append([text, activity_name, category_name])

    def __getattr__(self, name):
        return getattr(self.widget, name)


class CategoryEntry():
    """Category entry widget.

    widget (gtk.Entry): the associated category entry
    """
    def __init__(self, widget=None, **kwds):
        # widget and completion are already defined
        # e.g. in the glade edit_activity.ui file
        self.widget = widget
        if not self.widget:
            self.widget = gtk.Entry(**kwds)

        self.completion = self.widget.get_completion()
        if not self.completion:
            self.completion = gtk.EntryCompletion()
            self.widget.set_completion(self.completion)
        self.completion.insert_action_markup(0, "<i>Clear ({})</i>".format(_("Unsorted")))
        self.unsorted_action_index = 0

        self.model = gtk.ListStore(str)
        self.completion.set_model(self.model)
        self.completion.set_text_column(0)
        self.completion.set_match_func(self.match_func, None)

        self.widget.connect("icon-release", self.on_icon_release)
        self.widget.connect("focus-in-event", self.on_focus_in_event)
        self.completion.connect("action_activated", self.on_action_activated)

    def clear(self, notify=True):
        self.widget.set_text("")
        if notify:
            self.emit("changed")

    def match_func(self, completion, key, iter, *user_data):
        if not key.strip():
            # show all keys if entry is empty
            return True
        else:
            # return whether the entered string is
            # anywhere in the first column data
            return key.strip() in self.model.get_value(iter, 0).lower()

    def on_action_activated(self, completion, index):
        if index == self.unsorted_action_index:
            self.clear(notify=False)

    def on_focus_in_event(self, widget, event):
        self.populate_completions()

    def on_icon_release(self, entry, icon_pos, event):
        self.widget.grab_focus()
        # do not emit changed on the primary (clear) button
        self.clear()

    def populate_completions(self):
        self.model.clear()
        for category in runtime.storage.get_categories():
            self.model.append([category['name']])

    def __getattr__(self, name):
        return getattr(self.widget, name)



# - coding: utf-8 -

# Copyright (C) 2009 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject as gobject
from gi.repository import Gdk as gdk
from gi.repository import Gtk as gtk
from gi.repository import Pango as pango
import cairo
from math import pi

from hamster.lib import graphics, stuff
from hamster.lib.configuration import runtime

class TagsEntry(gtk.Entry):
    __gsignals__ = {
        'tags-selected': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }

    def __init__(self, *, parent):
        gtk.Entry.__init__(self, parent=parent)
        self.ac_tags = None  # "autocomplete" tags
        self.filter = None # currently applied filter string
        self.filter_tags = [] #filtered tags

        self.popup = gtk.Window(type = gtk.WindowType.POPUP)
        self.popup.set_attached_to(self)
        self.popup.set_transient_for(self.get_ancestor(gtk.Window))

        self.scroll_box = gtk.ScrolledWindow()
        self.scroll_box.set_shadow_type(gtk.ShadowType.IN)
        self.scroll_box.set_policy(gtk.PolicyType.NEVER, gtk.PolicyType.AUTOMATIC)
        viewport = gtk.Viewport()
        viewport.set_shadow_type(gtk.ShadowType.NONE)

        self.tag_box = TagBox()
        self.tag_box.connect("tag-selected", self.on_tag_selected)
        self.tag_box.connect("tag-unselected", self.on_tag_unselected)


        viewport.add(self.tag_box)
        self.scroll_box.add(viewport)
        self.popup.add(self.scroll_box)

        self.set_icon_from_icon_name(gtk.EntryIconPosition.SECONDARY, "go-down-symbolic")

        self.connect("icon-press", self._on_icon_press)
        self.connect("key-press-event", self._on_key_press_event)
        self.connect("focus-out-event", self._on_focus_out_event)

        self._parent_click_watcher = None # bit lame but works

        self.external_listeners = [
            (runtime.storage, runtime.storage.connect('tags-changed', self.refresh_ac_tags))
        ]
        self.show()
        self.populate_suggestions()
        self.connect("destroy", self.on_destroy)

    def on_destroy(self, window):
        for obj, handler in self.external_listeners:
            obj.disconnect(handler)
        self.popup.destroy()
        self.popup = None


    def refresh_ac_tags(self, event):
        self.ac_tags = None

    def get_tags(self):
        # splits the string by comma and filters out blanks
        return [tag.strip() for tag in self.get_text().split(",") if tag.strip()]

    def on_tag_selected(self, tag_box, tag):
        cursor_tag = self.get_cursor_tag()
        if cursor_tag and tag.lower().startswith(cursor_tag.lower()):
            self.replace_tag(cursor_tag, tag)
            tags = self.get_tags()
        else:
            tags = self.get_tags()
            tags.append(tag)

        self.tag_box.selected_tags = tags


        self.set_tags(tags)
        self.update_tagsline(add=True)

        self.populate_suggestions()
        self.show_popup()

    def on_tag_unselected(self, tag_box, tag):
        tags = self.get_tags()
        while tag in tags: #it could be that dear user is mocking us and entering same tag over and over again
            tags.remove(tag)

        self.tag_box.selected_tags = tags

        self.set_tags(tags)
        self.update_tagsline(add=True)

    def hide_popup(self):
        self.popup.hide()
        if self._parent_click_watcher and self.get_toplevel().handler_is_connected(self._parent_click_watcher):
            self.get_toplevel().disconnect(self._parent_click_watcher)
            self._parent_click_watcher = None

    def show_popup(self):
        if not self.filter_tags:
            self.popup.hide()
            return

        if not self._parent_click_watcher:
            self._parent_click_watcher = self.get_toplevel().connect("button-press-event", self._on_focus_out_event)

        alloc = self.get_allocation()
        _, x, y = self.get_parent_window().get_origin()

        self.popup.move(x + alloc.x,y + alloc.y + alloc.height)

        w = alloc.width

        height = self.tag_box.count_height(w)

        self.scroll_box.set_size_request(w, height)
        self.popup.resize(w, height)
        self.popup.show_all()

    def refresh_activities(self):
        # scratch activities and categories so that they get repopulated on demand
        self.activities = None
        self.categories = None

    def populate_suggestions(self):
        self.ac_tags = self.ac_tags or [tag["name"] for tag in
                                        runtime.storage.get_tags(only_autocomplete=True)]

        cursor_tag = self.get_cursor_tag()

        self.filter = cursor_tag

        entered_tags = self.get_tags()
        self.tag_box.selected_tags = entered_tags

        self.filter_tags = [tag for tag in self.ac_tags
                            if (tag or "").lower().startswith((self.filter or "").lower())]

        self.tag_box.draw(self.filter_tags)



    def _on_focus_out_event(self, widget, event):
        self.hide_popup()

    def _on_icon_press(self, entry, icon_pos, event):
        # otherwise Esc could not hide popup
        self.grab_focus()
        # toggle popup
        if self.popup.get_visible():
            # remove trailing comma if any
            self.update_tagsline(add=False)
            self.hide_popup()
        else:
            # add trailing comma
            self.update_tagsline(add=True)
            self.populate_suggestions()
            self.show_popup()

    def get_cursor_tag(self):
        #returns the tag on which the cursor is on right now
        if self.get_selection_bounds():
            cursor = self.get_selection_bounds()[0]
        else:
            cursor = self.get_position()

        text = self.get_text()

        return text[text.rfind(",", 0, cursor)+1:max(text.find(",", cursor+1)+1, len(text))].strip()


    def replace_tag(self, old_tag, new_tag):
        tags = self.get_tags()
        if old_tag in tags:
            tags[tags.index(old_tag)] = new_tag

        if self.get_selection_bounds():
            cursor = self.get_selection_bounds()[0]
        else:
            cursor = self.get_position()

        self.set_tags(tags)
        self.set_position(len(self.get_text()))

    def set_tags(self, tags):
        self.tags = tags
        self.update_tagsline()

    def update_tagsline(self, add=False):
        """Update tags line text.

        If add is True, prepare to add tags to the list:
        a comma is appended and the popup is displayed.
        """
        text = ", ".join(self.tags)
        if add and text:
            text = "{}, ".format(text)
        self.set_text(text)
        self.set_position(len(self.get_text()))

    def _on_key_press_event(self, entry, event):
        if event.keyval == gdk.KEY_Tab:
            if self.popup.get_property("visible"):
                #we have to replace
                if self.get_text() and self.get_cursor_tag() != self.filter_tags[0]:
                    self.replace_tag(self.get_cursor_tag(), self.filter_tags[0])
                    return True
                else:
                    return False
            else:
                return False

        elif event.keyval in (gdk.KEY_Return, gdk.KEY_KP_Enter):
            if self.popup.get_property("visible"):
                if self.get_text():
                    self.hide_popup()
                return True
            else:
                if self.get_text():
                    self.emit("tags-selected")
                return False

        elif event.keyval == gdk.KEY_Escape:
            if self.popup.get_property("visible"):
                self.hide_popup()
                return True
            else:
                return False

        else:
            self.populate_suggestions()
            self.show_popup()

        return False


class TagBox(graphics.Scene):
    __gsignals__ = {
        'tag-selected': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str,)),
        'tag-unselected': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str,)),
    }

    def __init__(self, interactive = True):
        graphics.Scene.__init__(self)
        self.interactive = interactive
        self.hover_tag = None
        self.tags = []
        self.selected_tags = []
        self.layout = None

        if self.interactive:
            self.connect("on-mouse-over", self.on_mouse_over)
            self.connect("on-mouse-out", self.on_mouse_out)
            self.connect("on-click", self.on_tag_click)

        self.connect("on-enter-frame", self.on_enter_frame)

    def on_mouse_over(self, area, tag):
        tag.color = tag.graphics.colors.darker(tag.color, -20)

    def on_mouse_out(self, area, tag):
        if tag.text in self.selected_tags:
            tag.color = (242, 229, 97)
        else:
            tag.color = (241, 234, 170)


    def on_tag_click(self, area, event, tag):
        if not tag: return

        if tag.text in self.selected_tags:
            self.emit("tag-unselected", tag.text)
        else:
            self.emit("tag-selected", tag.text)
        self.on_mouse_out(area, tag) #paint
        self.redraw()

    def draw(self, tags):
        new_tags = []
        for label in tags:
            tag = Tag(label)
            if label in self.selected_tags:
                tag.color = (242, 229, 97)
            new_tags.append(tag)

        for tag in self.tags:
            self.sprites.remove(tag)

        self.add_child(*new_tags)
        self.tags = new_tags

        self.show()
        self.redraw()

    def count_height(self, width):
        # reposition tags and see how much space we take up
        self.width = width
        w, h = self.on_enter_frame(None, None)
        return h + 6

    def on_enter_frame(self, scene, context):
        cur_x, cur_y = 4, 4
        tag = None
        for tag in self.tags:
            if cur_x + tag.width >= self.width - 5:  #if we do not fit, we wrap
                cur_x = 5
                cur_y += tag.height + 6

            tag.x = cur_x
            tag.y = cur_y

            cur_x += tag.width + 6 #some padding too, please

        if tag:
            cur_y += tag.height + 2 # the last one

        return cur_x, cur_y

class Tag(graphics.Sprite):
    def __init__(self, text, interactive = True, color = "#F1EAAA"):
        graphics.Sprite.__init__(self, interactive = interactive)

        self.width, self.height = 0,0

        font = gtk.Style().font_desc
        font_size = int(font.get_size() * 0.8 / pango.SCALE) # 80% of default

        self.label = graphics.Label(text, size = font_size, color = (30, 30, 30), y = 1)
        self.color = color
        self.add_child(self.label)

        self.corner = int((self.label.height + 3) / 3) + 0.5
        self.label.x = self.corner + 6

        self.text = stuff.escape_pango(text)
        self.connect("on-render", self.on_render)

    def __setattr__(self, name, value):
        graphics.Sprite.__setattr__(self, name, value)
        if name == 'text' and hasattr(self, 'label'):
            self.label.text = value
            self.__dict__['width'], self.__dict__['height'] = int(self.label.x + self.label.width + self.label.height * 0.3), self.label.height + 3

    def on_render(self, sprite):
        self.graphics.set_line_style(width=1)

        self.graphics.move_to(0.5, self.corner)
        self.graphics.line_to([(self.corner, 0.5),
                               (self.width + 0.5, 0.5),
                               (self.width + 0.5, self.height - 0.5),
                               (self.corner, self.height - 0.5),
                               (0.5, self.height - self.corner)])
        self.graphics.close_path()
        self.graphics.fill_stroke(self.color, "#b4b4b4")

        self.graphics.circle(6, self.height / 2, 2)
        self.graphics.fill_stroke("#fff", "#b4b4b4")



# This file is part of Hamster
# Copyright (c) The Hamster time tracker developers
# SPDX-License-Identifier: GPL-3.0-or-later


"""Hamster datetime.

Python datetime replacement, tuned for hamster use.
"""


import logging
logger = logging.getLogger(__name__)   # noqa: E402

import datetime as pdt  # standard datetime
import re

from collections import namedtuple
from textwrap import dedent
from functools import lru_cache


class datetime:  # predeclaration for return type annotations
    pass


class time:  # predeclaration for return type annotations
    pass


class date(pdt.date):
    """Hamster date.

    Should replace the python datetime.date in any customer code.

    Same as python date, with the addition of parse methods.
    """

    FMT = "%Y-%m-%d"  # ISO format

    def __new__(cls, year, month, day):
        return pdt.date.__new__(cls, year, month, day)

    def __add__(self, other):
        # python date.__add__ was not type stable prior to 3.8
        return self.from_pdt(self.to_pdt() + other)

    __radd__ = __add__

    def __sub__(self, other):
        # python date.__sub__ was not type stable prior to 3.8
        if isinstance(other, pdt.timedelta):
            return self.from_pdt(self.to_pdt() - other)
        elif isinstance(other, pdt.date):
            return timedelta.from_pdt(self.to_pdt() - other)
        else:
            return NotImplemented

    @classmethod
    def parse(cls, s):
        """Return date from string."""
        m = cls.re.search(s)
        return cls(year=int(m.group('year')),
                   month=int(m.group('month')),
                   day=int(m.group('day'))
                   )

    @classmethod
    def pattern(cls, detailed=True):
        if detailed:
            return dedent(r"""
                (?P<year>\d{4})        # 4 digits
                -                      # dash
                (?P<month>\d{2})       # 2 digits
                -                      # dash
                (?P<day>\d{2})         # 2 digits
                """)
        else:
            return r"""\d{4}-\d{2}-\d{2}"""

    @classmethod
    def from_pdt(cls, d):
        """Convert python date to hamster date."""
        return cls(d.year, d.month, d.day)

    def to_pdt(self):
        """Convert to python date."""
        return pdt.date(self.year, self.month, self.day)

# For datetime that will need to be outside the class.
# Same here for consistency
date.re = re.compile(date.pattern(), flags=re.VERBOSE)


class hday(date):
    """Hamster day.

    Same as date, but taking into account day-start.
    A day starts at conf.day_start and ends when the next day starts.
    """

    def __new__(cls, year, month, day):
        return date.__new__(cls, year, month, day)

    @property
    def end(self) -> datetime:
        """Day end."""
        return datetime.from_day_time(self + timedelta(days=1), self.start_time())

    @property
    def start(self) -> datetime:
        """Day start."""
        return datetime.from_day_time(self, self.start_time())

    @classmethod
    def start_time(cls) -> time:
        """Day start time."""
        # work around cyclic imports
        from hamster.lib.configuration import conf
        return conf.day_start

    @classmethod
    def today(cls):
        """Return the current hamster day."""
        return datetime.now().hday()


class time(pdt.time):
    """Hamster time.

    Should replace the python datetime.time in any customer code.
    Specificities:
    - rounded to minutes
    - conversion to and from string facilities
    """

    FMT = "%H:%M"  # display format, e.g. 13:30

    def __new__(cls,
                hour=0, minute=0,
                second=0, microsecond=0,
                tzinfo=None, **kwargs):
            # round down to zero seconds and microseconds
            return pdt.time.__new__(cls,
                                    hour=hour, minute=minute,
                                    second=0, microsecond=0,
                                    tzinfo=None, **kwargs)

    @classmethod
    def _extract_time(cls, match, h="hour", m="minute"):
        """Extract time from a time.re match.

        Custom group names allow to use the same method
        for two times in the same regexp (e.g. for range parsing)

        h (str): name of the group containing the hour
        m (str): name of the group containing the minute

        seealso: time.parse
        """
        h_str = match.group(h)
        m_str = match.group(m)
        if h_str and m_str:
            hour = int(h_str)
            minute = int(m_str)
            return cls(hour, minute)
        else:
            return None

    @classmethod
    def parse(cls, s):
        """Parse time from string."""
        m = cls.re.search(s)
        return cls._extract_time(m) if m else None

    # For datetime that must be a method.
    # Same here for consistency.
    @classmethod
    def pattern(cls):
        """Return a time pattern with all group names."""

        # remove the indentation for easier debugging.
        return dedent(r"""
            (?P<hour>                         # hour
             [0-9](?=[,\.:])                  # positive lookahead:
                                              # allow a single digit only if
                                              # followed by a colon, dot or comma
             | [0-1][0-9]                     # 00 to 19
             | [2][0-3]                       # 20 to 23
            )
            [,\.:]?                           # Separator can be colon,
                                              # dot, comma, or nothing.
            (?P<minute>[0-5][0-9])            # minute (2 digits, between 00 and 59)
            (?!\d?-\d{2}-\d{2})               # Negative lookahead:
                                              # avoid matching date by inadvertance.
                                              # For instance 2019-12-05
                                              # might be caught as 2:01.
                                              # Requiring space or - would not work:
                                              # 2019-2025 is the 20:19-20:25 range.
            """)


# For datetime that will need to be outside the class.
# Same here for consistency
time.re = re.compile(time.pattern(), flags=re.VERBOSE)


class datetime(pdt.datetime):
    """Hamster datetime.

    Should replace the python datetime.datetime in any customer code.
    Specificities:
    - rounded to minutes
    - conversion to and from string facilities
    """

    # display format, e.g. 2020-01-20 20:40
    FMT = "{} {}".format(date.FMT, time.FMT)

    def __new__(cls, year, month, day,
                hour=0, minute=0,
                second=0, microsecond=0,
                tzinfo=None, **kwargs):
            # round down to zero seconds and microseconds
            return pdt.datetime.__new__(cls, year, month, day,
                                        hour=hour, minute=minute,
                                        second=0, microsecond=0,
                                        tzinfo=None, **kwargs)

    def __add__(self, other):
        # python datetime.__add__ was not type stable prior to 3.8
        return datetime.from_pdt(self.to_pdt() + other)

    # similar to https://stackoverflow.com/q/51966126/3565696
    # __getnewargs_ex__ did not work, brute force required
    def __deepcopy__(self, memo):
        kwargs = {}
        if (hasattr(self, 'fold')):
            kwargs['fold'] = self.fold
        return datetime(self.year, self.month, self.day,
                        self.hour, self.minute,
                        self.second, self.microsecond,
                        self.tzinfo, **kwargs)
    __radd__ = __add__

    def __sub__(self, other):
        # python datetime.__sub__ was not type stable prior to 3.8
        if isinstance(other, timedelta):
            return datetime.from_pdt(self.to_pdt() - other)
        elif isinstance(other, datetime):
            return timedelta.from_pdt(self.to_pdt() - other)
        else:
            return NotImplemented

    def __str__(self):
        if self.tzinfo:
            raise NotImplementedError("Stay tuned...")
        else:
            return self.strftime(self.FMT)

    @classmethod
    def _extract_datetime(cls, match, d="date", h="hour", m="minute", r="relative",
                          default_day=None):
        """extract datetime from a datetime.pattern match.

        Custom group names allow to use the same method
        for two datetimes in the same regexp (e.g. for range parsing)

        h (str): name of the group containing the hour
        m (str): name of the group containing the minute
        r (str): name of the group containing the relative time
        default_day (dt.date): the datetime will belong to this hamster day if
                               date is missing.
        """
        _time = time._extract_time(match, h, m)
        if _time:
            date_str = match.group(d)
            if date_str:
                _date = date.parse(date_str)
                return datetime.combine(_date, _time)
            else:
                return datetime.from_day_time(default_day, _time)
        else:
            relative_str = match.group(r)
            if relative_str and relative_str != "--":
                return timedelta(minutes=int(relative_str))
            else:
                return None

    # not a property, to match other extractions such as .date() or .time()
    def hday(self) -> hday:
        """Return the day belonged to.

        The hamster day start is taken into account.
        """

        # work around cyclic imports
        from hamster.lib.configuration import conf

        _day = hday(self.year, self.month, self.day)
        if self.time() < conf.day_start:
            # early morning, between midnight and day_start
            # => the hamster day is the previous civil day
            _day -= timedelta(days=1)

        # return only the date
        return _day

    @classmethod
    def from_day_time(cls, d: hday, t: time):
        """Return a datetime with time t belonging to day.

        The hamster day start is taken into account.
        """

        # work around cyclic imports
        from hamster.lib.configuration import conf

        if t < conf.day_start:
            # early morning, between midnight and day_start
            # => the hamster day is the previous civil day
            civil_date = d + timedelta(days=1)
        else:
            civil_date = d
        return cls.combine(civil_date, t)

    # Note: fromisoformat appears in python3.7

    @classmethod
    def from_pdt(cls, t):
        """Convert python datetime to hamster datetime."""
        kwargs = {}
        if (hasattr(t, 'fold')):
            kwargs['fold'] = t.fold
        return cls(t.year, t.month, t.day,
                   t.hour, t.minute,
                   t.second, t.microsecond,
                   t.tzinfo, **kwargs)

    @classmethod
    def now(cls):
        """Current datetime."""
        return cls.from_pdt(pdt.datetime.now())

    @classmethod
    def parse(cls, s, default_day=None):
        """Parse a datetime from text.

        default_day (dt.date):
            If start is given without any date (e.g. just hh:mm),
            put the corresponding datetime in default_day.
            Defaults to today.
        """

        # datetime.re is added below, after the class definition
        # it will be found at runtime
        m = datetime.re.search(s)
        return cls._extract_datetime(m, default_day=default_day) if m else None

    @classmethod
    @lru_cache()
    def pattern(cls, n=None):
        """Return a datetime pattern with all group names.

        If n is given, all groups are suffixed with str(n).
        """

        # remove the indentation => easier debugging.
        base_pattern = dedent(r"""
            (?P<whole>
                                              # note: need to double the brackets
                                              #       for .format
                (?<!\d{{2}})                  # negative lookbehind,
                                              # avoid matching 2019-12 or 2019-12-05
                (?P<relative>
                    --                        # double dash: None
                   |                          # or
                    [-+]                      # minus or plus: relative to ref
                    \d{{1,3}}                 # 1, 2 or 3 digits
                )
            |                             # or
                (?P<date>{})?                 # maybe date
                \s?                           # maybe one space
                {}                            # time
            )
            """).format(date.pattern(), time.pattern())
        if n is None:
            return base_pattern
        else:
            to_replace = ("whole", "relative",
                          "year", "month", "day", "date", "tens", "hour", "minute")
            specifics = ["{}{}".format(s, n) for s in to_replace]
            res = base_pattern
            for src, dest in zip(to_replace, specifics):
                res = res.replace(src, dest)
            return res

    def to_pdt(self):
        """Convert to python datetime."""
        kwargs = {}
        if (hasattr(self, 'fold')):
            kwargs['fold'] = self.fold
        return pdt.datetime(self.year, self.month, self.day,
                            self.hour, self.minute,
                            self.second, self.microsecond,
                            self.tzinfo, **kwargs)


# outside class; need the class to be defined first
datetime.re = re.compile(datetime.pattern(), flags=re.VERBOSE)


class Range():
    """Time span between two datetimes."""

    # slight memory optimization; no further attributes besides start or end.
    __slots__ = ('start', 'end')

    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end

    def __bool__(self):
        return not (self.start is None and self.end is None)

    def __eq__(self, other):
        if isinstance(other, Range):
            return self.start == other.start and self.end == other.end
        else:
            return False

    # allow start, end = range
    def __iter__(self):
        return (self.start, self.end).__iter__()

    def format(self, default_day=None, explicit_none=True):
        """Return a string representing the time range.

        Start date is shown only if start does not belong to default_day.
        End date is shown only if end does not belong to
        the same hamster day as start (or to default_day if start is None).
        """

        none_str = "--" if explicit_none else ""

        if self.start:
            if self.start.hday() != default_day:
                start_str = self.start.strftime(datetime.FMT)
            else:
                start_str = self.start.strftime(time.FMT)
            default_end_day = self.start.hday()
        else:
            start_str = none_str
            default_end_day = default_day

        if self.end:
            if self.end.hday() != default_end_day:
                end_str = self.end.strftime(datetime.FMT)
            else:
                end_str = self.end.strftime(time.FMT)
        else:
            end_str = none_str

        if end_str:
            return "{} - {}".format(start_str, end_str)
        else:
            return start_str

    @classmethod
    def parse(cls, text,
              position="exact", separator="\s+", default_day=None, ref="now"):
        """Parse a start-end range from text.

        position (str): "exact" to match exactly the full text
                        "head" to search only at the beginning of text, and
                        "tail" to search only at the end.

        separator (str): regexp pattern (e.g. '\s+') meant to separate the datetime
                         from the rest. Discarded for "exact" position.

        default_day (date): If start is given without any date (e.g. just hh:mm),
                               put the corresponding datetime in default_day.
                               Defaults to today.
                               Note: the default end day is always the start day, so
                                     "2019-11-27 23:50 - 00:20" lasts 30 minutes.

        ref (datetime): reference for relative times
                           (e.g. -15: quarter hour before ref).
                           For testing purposes only
                           (note: this will be removed later on,
                            and replaced with datetime.now mocking in pytest).
                           For users, it should be "now".
        Return:
            (range, rest)
            range (Range): Range(None, None) if no match is found.
            rest (str): remainder of the text.
        """

        if ref == "now":
            ref = datetime.now()

        if default_day is None:
            default_day = hday.today()

        assert position in ("exact", "head", "tail"), "position unknown: '{}'".format(position)
        if position == "exact":
            p = "^{}$".format(cls.pattern())
        elif position == "head":
            # ( )?: require either only the range (no rest),
            #       or separator between range and rest,
            #       to avoid matching 10.00@cat
            # .*? so rest is as little as possible
            p = "^{}( {}(?P<rest>.*?) )?$".format(cls.pattern(), separator)
        elif position == "tail":
            p = "^( (?P<rest>.*?){} )? {}$".format(separator, cls.pattern())
        # No need to compile, recent patterns are cached by re.
        # DOTALL, so rest may contain newlines
        # (important for multiline descriptions)
        m = re.search(p, text, flags=re.VERBOSE | re.DOTALL)

        if not m:
            return Range(None, None), text
        elif position == "exact":
            rest = ""
        else:
            rest = m.group("rest") or ""

        if m.group('firstday'):
            # only day given for start
            firstday = hday.parse(m.group('firstday'))
            start = firstday.start
        else:
            firstday = None
            start = datetime._extract_datetime(m, d="date1", h="hour1",
                                               m="minute1", r="relative1",
                                               default_day=default_day)
            if isinstance(start, pdt.timedelta):
                # relative to ref, actually
                assert ref, "relative start needs ref"
                start = ref + start

        if m.group('lastday'):
            lastday = hday.parse(m.group('lastday'))
            end = lastday.end
        elif firstday:
            end = firstday.end
        elif m.group('duration'):
            duration = int(m.group('duration'))
            end = start + timedelta(minutes=duration)
        else:
            end_default_day = start.hday() if start else default_day
            end = datetime._extract_datetime(m, d="date2", h="hour2",
                                             m="minute2", r="relative2",
                                             default_day=end_default_day)
            if isinstance(end, pdt.timedelta):
                # relative to ref, actually
                assert ref, "relative end needs ref"
                end = ref + end

        return Range(start, end), rest

    @classmethod
    @lru_cache()
    def pattern(cls):
        return dedent(r"""
            (                    # start
              {}                 # datetime: relative1 or (date1, hour1, and minute1)
              |                  # or
              (?P<firstday>{})  # date without time
            )
            (

                (?P<separation>   # (only needed if end time is given)
                    \s?           # maybe one space
                    -             # dash
                    \s?           # maybe one space
                  |               # or
                    \s            # one space exactly
                )
            (                     # end
              {}                  # datetime: relative2 or (date2, hour2, and minute2)
              |                   # or
              (?P<lastday>{})     # date without time
              |
              (?P<duration>
                  (?<![+-])       # negative lookbehind: no sign
                  \d{{1,3}}       # 1, 2 or 3 digits
              )
            )
            )?                    # end time is facultative
            """.format(datetime.pattern(1), date.pattern(detailed=False),
                       datetime.pattern(2), date.pattern(detailed=False)))

    @classmethod
    def from_start_end(cls, start, end=None):
        """Return a Range from separate arguments.

        Convenient to ease backward compatibility,
        and to handle either hdays or datetimes.
        """
        if isinstance(start, Range):
            assert end is None, "end cannot be passed together with a Range"
            return cls(start.start, start.end)
        elif (start is None) or isinstance(start, datetime):
            # This one must come first,
            # because inheritance order is datetime < pdt.datetime < pdt.date.
            pass
        elif isinstance(start, hday):
            # Idem, beware of the inheritance order;
            # hday < date < pdt.date.
            day = start
            start = day.start
            if end is None:
                end = day.end
        elif isinstance(start, pdt.date):
            # transition from legacy
            start = hday.from_pdt(start).start
        else:
            raise TypeError(
                "\n    First argument should be either Range, None, datetime or hday;"
                "\n    received {}".format(type(start))
                )

        if (end is None) or isinstance(end, datetime):
            # same as above
            pass
        elif isinstance(end, hday):
            end = end.end
        elif isinstance(end, pdt.date):
            # transition from legacy
            end = hday.from_pdt(end).end
        else:
            raise TypeError(
                "\n    Second argument should be either None, datetime or hday;"
                "\n    received {}".format(type(start)))

        return cls(start, end)

    @classmethod
    def today(cls):
        _today = hday.today()
        return cls(_today.start, _today.end)


class timedelta(pdt.timedelta):
    """Hamster timedelta.

    Should replace the python datetime.timedelta in any customer code.
    Specificities:
    - conversion to and from string facilities
    Note: granularity is the same as the original python timedelta
          (only datetimes should be truncated).
    """

    def __new__(cls, days=0, seconds=0, microseconds=0,
                milliseconds=0, minutes=0, hours=0, weeks=0):
        # Tempted to round down ? Resist. Not so useful + issues down the line.
        return pdt.timedelta.__new__(cls,
                                     days=days,
                                     seconds=seconds,
                                     microseconds=microseconds,
                                     milliseconds=milliseconds,
                                     minutes=minutes,
                                     hours=hours,
                                     weeks=weeks)

    # timedelta subclassing is not type stable yet
    def __add__(self, other):
        return timedelta.from_pdt(self.to_pdt() + other)

    __radd__ = __add__

    def __sub__(self, other):
        return timedelta.from_pdt(self.to_pdt() - other)

    def __neg__(self):
        return timedelta.from_pdt(-self.to_pdt())

    @classmethod
    def from_pdt(cls, delta):
        """Convert python timedelta to hamster timedelta."""

        # Only days, seconds and microseconds are stored internally
        return cls(days=delta.days,
                   seconds=delta.seconds,
                   microseconds=delta.microseconds)

    def to_pdt(self):
        """Convert to python timedelta."""

        return pdt.timedelta(days=self.days,
                             seconds=self.seconds,
                             microseconds=self.microseconds)

    def format(self, fmt="human"):
        """Return a string representation, according to the format string fmt."""
        allowed = ("human", "HH:MM")

        # TODO: should use total_minutes
        total_s = self.total_seconds()
        if total_s < 0:
            # return a warning
            return "NEGATIVE"

        hours = int(total_s // 3600)
        minutes = int((total_s - 3600 * hours) // 60)

        if fmt == "human":
            if minutes % 60 == 0:
                # duration in round hours
                return "{}h".format(hours)
            elif hours == 0:
                # duration less than one hour
                return "{}min".format(minutes)
            else:
                # x hours, y minutes
                return "{}h {}min".format(hours, minutes)
        elif fmt == "HH:MM":
            return "{:02d}:{:02d}".format(hours, minutes)
        else:
            raise NotImplementedError(
                "'{}' not in allowed formats: {}".format(fmt, allowed))

    def total_minutes(self):
        """Return the duration in minutes (float)."""
        return self.total_seconds() / 60



# Should not normally be used directly, use hamster.__version__ and
# hamster.installed instead

def get_installed_version():
    try:
        # defs.py is created by waf from defs.py.in
        from hamster import defs
        return defs.VERSION
    except ImportError:
        # if defs is not there, we are running from sources
        return None

def get_uninstalled_version():
    # If available, prefer the git version, otherwise fall back to
    # the VERSION file (which is meaningful only in released
    # versions)
    from subprocess import getstatusoutput
    rc, output = getstatusoutput("git describe --tags --always --dirty=+")
    if rc == 0:
        import re
        # Strip "v" prefix that is used in git tags
        return re.sub(r'^v', '', output)
    else:
        from pathlib import Path
        with open(Path(__file__).parent / 'VERSION', 'r') as f:
            return f.read()

def get_version():
    """
    Figure out the hamster version.

    Returns a tuple with the version string and wether we are installed or not.
    """

    version = get_installed_version()
    if version is not None:
        return (version, True)

    version = get_uninstalled_version()
    return ("{} (uninstalled)".format(version), False)


if __name__ == '__main__':
    import sys
    # Intended to be called by waf when installing, so only return
    # uninstalled version
    sys.stdout.write(get_uninstalled_version())



# - coding: utf-8 -

# Copyright (C) 2008-2009, 2014 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject as gobject
from gi.repository import Gtk as gtk
from gi.repository import Pango as pango

import calendar
import re

from hamster.lib import datetime as dt
from hamster.lib import stuff
from hamster.lib.configuration import load_ui_file


class Calendar():
    """Python date interface to a Gtk.Calendar.

    widget (Gtk.Calendar):
        the associated Gtk widget.
    expander (Gtk.expander):
        An optional expander which contains the widget.
        The expander label displays the date.
    """
    def __init__(self, widget, expander=None):
        self.widget = widget
        self.expander = expander
        self.widget.connect("day-selected", self.on_date_changed)

    @property
    def date(self):
        """Selected day, as datetime.date."""
        year, month, day = self.widget.get_date()
        # months start at 0 in Gtk.Calendar and at 1 in python date
        month += 1
        return dt.date(year=year, month=month, day=day) if day else None

    @date.setter
    def date(self, value):
        """Set date.

        value can be a python date or datetime.
        """
        if value is None:
            # unselect day
            self.widget.select_day(0)
        else:
            year = value.year
            # months start at 0 in Gtk.Calendar and at 1 in python date
            month = value.month - 1
            day = value.day
            self.widget.select_month(month, year)
            self.widget.select_day(day)

    def on_date_changed(self, widget):
        if self.expander:
            if self.date:
                self.expander.set_label(self.date.strftime("%A %Y-%m-%d"))
            else:
                self.expander.set_label("")

    def __getattr__(self, name):
        return getattr(self.widget, name)


class RangePick(gtk.MenuButton):
    """ a text entry widget with calendar popup"""
    __gsignals__ = {
        # day|week|month|manual, start, end
        'range-selected': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
    }


    def __init__(self, today):
        gtk.ToggleButton.__init__(self)

        self._ui = load_ui_file("date_range.ui")

        popup = self.get_widget("range_popup")
        popup.connect("show", self.on_show)
        self.set_popover(popup)

        self.today = today

        hbox = gtk.HBox()
        hbox.set_spacing(3)
        self.label = gtk.Label()
        hbox.add(self.label)
        hbox.add(gtk.Arrow(gtk.ArrowType.DOWN, gtk.ShadowType.ETCHED_IN))
        self.add(hbox)

        self.start_date, self.end_date = None, None
        self.current_range = None

        self._ui.connect_signals(self)

    def set_range(self, start_date, end_date=None):
        end_date = end_date or start_date
        self.start_date, self.end_date = start_date, end_date
        self.label.set_markup('<b>%s</b>' % stuff.format_range(start_date, end_date))

    def get_range(self):
        return self.start_date, self.end_date

    def emit_range(self, range, start, end):
        self.set_range(start, end)
        self.emit("range-selected", range, start, end)
        self.set_active(False)


    def prev_range(self):
        start, end = self.start_date, self.end_date

        if self.current_range == "day":
            start, end = start - dt.timedelta(1), end - dt.timedelta(1)
        elif self.current_range == "week":
            start, end = start - dt.timedelta(7), end - dt.timedelta(7)
        elif self.current_range == "month":
            end = start - dt.timedelta(1)
            first_weekday, days_in_month = calendar.monthrange(end.year, end.month)
            start = end - dt.timedelta(days_in_month - 1)
        else:
            # manual range - just jump to the next window
            days =  (end - start) + dt.timedelta(days = 1)
            start = start - days
            end = end - days
        self.emit_range(self.current_range, start, end)


    def next_range(self):
        start, end = self.start_date, self.end_date

        if self.current_range == "day":
            start, end = start + dt.timedelta(1), end + dt.timedelta(1)
        elif self.current_range == "week":
            start, end = start + dt.timedelta(7), end + dt.timedelta(7)
        elif self.current_range == "month":
            start = end + dt.timedelta(1)
            first_weekday, days_in_month = calendar.monthrange(start.year, start.month)
            end = start + dt.timedelta(days_in_month - 1)
        else:
            # manual range - just jump to the next window
            days =  (end - start) + dt.timedelta(days = 1)
            start = start + days
            end = end + days

        self.emit_range(self.current_range, start, end)



    def get_widget(self, name):
        """ skip one variable (huh) """
        return self._ui.get_object(name)


    def on_show(self, user_data):
        self.get_widget("day_preview").set_text(stuff.format_range(self.today, self.today))
        self.get_widget("week_preview").set_text(stuff.format_range(*stuff.week(self.today)))
        self.get_widget("month_preview").set_text(stuff.format_range(*stuff.month(self.today)))

        start_cal = self.get_widget("start_calendar")
        start_cal.select_month(self.start_date.month - 1, self.start_date.year)
        start_cal.select_day(self.start_date.day)

        end_cal = self.get_widget("end_calendar")
        end_cal.select_month(self.end_date.month - 1, self.end_date.year)
        end_cal.select_day(self.end_date.day)

        self.get_widget("day").grab_focus()


    def on_day_clicked(self, button):
        self.current_range = "day"
        self.emit_range("day", self.today, self.today)

    def on_week_clicked(self, button):
        self.current_range = "week"
        self.start_date, self.end_date = stuff.week(self.today)
        self.emit_range("week", self.start_date, self.end_date)

    def on_month_clicked(self, button):
        self.current_range = "month"
        self.start_date, self.end_date = stuff.month(self.today)
        self.emit_range("month", self.start_date, self.end_date)

    def on_manual_range_apply_clicked(self, button):
        self.current_range = "manual"
        # GtkCalendar January is 0, hence the + 1
        year, month, day = self.get_widget("start_calendar").get_date()
        self.start_date = dt.date(year, month + 1, day)

        year, month, day = self.get_widget("end_calendar").get_date()
        self.end_date = dt.date(year, month + 1, day)

        # make sure we always have a valid range
        if self.end_date < self.start_date:
            self.start_date, self.end_date = self.end_date, self.start_date

        self.emit_range("manual", self.start_date, self.end_date)



# -*- coding: utf-8 -*-

# Copyright (C) 2008-2009, 2014 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

import bisect
import cairo

from collections import defaultdict
from gi.repository import GObject as gobject
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import PangoCairo as pangocairo
from gi.repository import Pango as pango

from hamster.lib import datetime as dt
from hamster.lib import graphics
from hamster.lib import stuff
from hamster.lib.fact import Fact


class ActionRow(graphics.Sprite):
    def __init__(self):
        graphics.Sprite.__init__(self)
        self.visible = False

        self.restart = graphics.Icon("view-refresh-symbolic", size=18,
                                     interactive=True,
                                     mouse_cursor=gdk.CursorType.HAND1,
                                     y=4)
        self.add_child(self.restart)

        self.width = 50  # Simon says


class TotalFact(Fact):
    """An extension of Fact that is used for daily totals.
    Instances of this class are rendered differently than instances
    of Fact.
    A TotalFact doesn't have a meaningful  start and an end, but a
    total duration (delta).
    FIXME: Ideally, we should have a common parent for Fact and Total Fact
    so we don't need to have nonsensical start and end properties here.
    """

    def __init__(self, activity, duration):
        super().__init__(activity=activity, start=dt.datetime.now(), end=dt.datetime.now())
        self.duration = duration

    @property
    def delta(self):
        return self.duration


class Label(object):
    """a much cheaper label that would be suitable for cellrenderer"""

    def __init__(self, x=0, y=0, color=None):
        self.x = x
        self.y = y
        self.color = color
        self._label_context = cairo.Context(cairo.ImageSurface(cairo.FORMAT_A1, 0, 0))
        self.layout = pangocairo.create_layout(self._label_context)
        self.layout.set_font_description(pango.FontDescription(graphics._font_desc))
        self.set_text("00:00 - 00:00")  # dummy time_label for finding width

    @property
    def height(self):
        """Label height in pixels."""
        return self.layout.get_pixel_size()[1]

    def set_text(self, text):
        self.text = text
        self.layout.set_markup(text)

    def get_text(self):
        return self.text

    def show(self, g, text=None, x=None, y=None):
        """Show the label.

        If text is given, it overrides any previous set_text().
        x and y can be passed to temporary override the position.
        (self.x and self.y will not be changed)
        """

        g.save_context()

        # fallback to self.x
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        g.move_to(x, y)

        if text is not None:
            self.set_text(text)

        if self.color:
            g.set_color(self.color)
        pangocairo.show_layout(g.context, self.layout)

        g.restore_context()


class TagLabel(Label):
    """Tag label, with small text."""

    def set_text(self, text):
        Label.set_text(self, "<small>{}</small>".format(text))


class FactRow(object):
    def __init__(self):
        self.time_label = Label()
        x = self.time_label.layout.get_pixel_size()[0] + 10
        self.activity_label = Label(x=x)

        self.category_label = Label()
        self.description_label = Label()
        self.tag_label = TagLabel()

        self.duration_label = Label()
        self.duration_label.layout.set_alignment(pango.Alignment.RIGHT)
        self.duration_label.layout.set_width(90 * pango.SCALE)

        self.width = 0

        # margins (in pixels)
        self.tag_row_margin_H = 2.5
        self.tag_row_margin_V = 2.5
        self.tag_inner_margin_H = 3
        self.tag_inner_margin_V = 2
        self.inter_tag_margin = 4
        self.row_margin_H = 5
        self.row_margin_V = 2
        self.category_offset_V = self.category_label.height * 0.1

    @property
    def height(self):
        res = self.activity_label.height + 2 * 3
        if self.fact.description:
            res += self.description_label.height

        if self.fact.tags:
            res += (self.tag_label.height
                    + self.tag_inner_margin_V * 2
                    + self.tag_row_margin_V * 2)

        res += self.row_margin_V * 2

        return res

    def set_fact(self, fact):
        """Set current fact."""

        self.fact = fact

        time_label = fact.start_time.strftime("%H:%M -")
        if fact.end_time:
            time_label += fact.end_time.strftime(" %H:%M")
        self.time_label.set_text(time_label)

        self.activity_label.set_text(stuff.escape_pango(fact.activity))

        category_text = "  - {}".format(stuff.escape_pango(fact.category)) if fact.category else ""
        self.category_label.set_text(category_text)

        text = stuff.escape_pango(fact.description)
        description_text = "<small><i>{}</i></small>".format(text) if fact.description else ""
        self.description_label.set_text(description_text)

        if fact.tags:
            # for now, tags are on a single line.
            # The first one is enough to determine the height.
            self.tag_label.set_text(stuff.escape_pango(fact.tags[0]))

    def _show_tags(self, g, color, bg):
        label = self.tag_label
        label.color = bg

        g.save_context()
        g.translate(self.tag_row_margin_H, self.tag_row_margin_V)
        for tag in self.fact.tags:
            label.set_text(stuff.escape_pango(tag))
            w, h = label.layout.get_pixel_size()
            rw = w + self.tag_inner_margin_H * 2
            rh = h + self.tag_inner_margin_V * 2
            g.rectangle(0, 0, rw, rh, 2)
            g.fill(color, 0.5)
            label.show(g, x=self.tag_inner_margin_H, y=self.tag_inner_margin_V)

            g.translate(rw + self.inter_tag_margin, 0)

        g.restore_context()

    def show(self, g, colors, fact=None, is_selected=False):
        """Display the fact row.

        If fact is given, the fact attribute is updated.
        """
        g.save_context()

        if fact is not None:
            # before the selection highlight, to get the correct height
            self.set_fact(fact)

        color, bg = colors["normal"], colors["normal_bg"]
        if is_selected:
            color, bg = colors["selected"], colors["selected_bg"]
            g.fill_area(0, 0, self.width, self.height, bg)

        g.translate(self.row_margin_H, self.row_margin_V)

        g.set_color(color)

        # Do not show the start/end time for Totals
        if not isinstance(self.fact, TotalFact):
            self.time_label.show(g)
        self.activity_label.show(g, self.activity_label.get_text() if not isinstance(self.fact, TotalFact) else "<b>{}</b>".format(self.activity_label.get_text()))

        if self.fact.category:
            g.save_context()
            category_color = graphics.ColorUtils.mix(bg, color, 0.57)
            g.set_color(category_color)
            x = self.activity_label.x + self.activity_label.layout.get_pixel_size()[0]
            self.category_label.show(g, x=x, y=self.category_offset_V)
            g.restore_context()

        if self.fact.description or self.fact.tags:
            g.save_context()
            g.translate(self.activity_label.x, self.activity_label.height + 3)

            if self.fact.tags:
                self._show_tags(g, color, bg)
                tag_height = (self.tag_label.height
                              + self.tag_inner_margin_V * 2
                              + self.tag_row_margin_V * 2)
                g.translate(0, tag_height)

            if self.fact.description:
                self.description_label.show(g)

            g.restore_context()

        self.duration_label.show(g, self.fact.delta.format() if not isinstance(self.fact, TotalFact) else "<b>{}</b>".format(self.fact.delta.format()), x=self.width - 105)

        g.restore_context()


class FactTree(graphics.Scene, gtk.Scrollable):
    """
    The fact tree is a painter.
    It does not change facts by itself, only sends signals.
    Facts get updated only through `set_facts`.

    It maintains scroll state and shows what we can see.
    That means it does not show all the facts there are,
    but rather only those that you can see.
    It's also painter as it reuses labels.
    Caching is futile, we do all the painting every time


    ASCII Art!
    | Weekday    | Start - End | Activity - category   [actions]| Duration       |
    | Month, Day |             | tags, description              |                |
    |            | Start - End | Activity - category            | Duration       |
    |            |             | Total                          | Total Duration |

    Inline edit?

    """

    __gsignals__ = {
        # enter or double-click, passes in current day and fact
        'on-activate-row': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        'on-delete-called': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,)),
    }

    hadjustment = gobject.property(type=gtk.Adjustment, default=None)
    hscroll_policy = gobject.property(type=gtk.ScrollablePolicy, default=gtk.ScrollablePolicy.MINIMUM)
    vadjustment = gobject.property(type=gtk.Adjustment, default=None)
    vscroll_policy = gobject.property(type=gtk.ScrollablePolicy, default=gtk.ScrollablePolicy.MINIMUM)

    def __init__(self):
        graphics.Scene.__init__(self, style_class=gtk.STYLE_CLASS_VIEW)

        self.date_label = Label(10, 3)
        fontdesc = pango.FontDescription(graphics._font_desc)
        fontdesc.set_weight(pango.Weight.BOLD)
        self.date_label.layout.set_alignment(pango.Alignment.RIGHT)
        self.date_label.layout.set_width(80 * pango.SCALE)
        self.date_label.layout.set_font_description(fontdesc)

        self.fact_row = FactRow()

        self.action_row = ActionRow()
        # self.add_child(self.action_row)

        self.row_positions = []
        self.row_heights = []

        self.y = 0
        self.day_padding = 20

        self.hover_day = None
        self.hover_fact = None
        self.current_fact = None

        self.style = self._style

        self.visible_range = None
        self.set_size_request(500, 400)

        self.connect("on-mouse-scroll", self.on_scroll)
        self.connect("on-mouse-move", self.on_mouse_move)
        self.connect("on-mouse-down", self.on_mouse_down)

        self.connect("on-resize", self.on_resize)
        self.connect("on-key-press", self.on_key_press)
        self.connect("notify::vadjustment", self._on_vadjustment_change)
        self.connect("on-enter-frame", self.on_enter_frame)
        self.connect("on-double-click", self.on_double_click)

    @property
    def current_fact_index(self):
        """Current fact index in the self.facts list."""
        facts_ids = [fact.id for fact in self.facts]
        return facts_ids.index(self.current_fact.id)

    def on_mouse_down(self, scene, event):
        self.on_mouse_move(None, event)
        self.grab_focus()
        if self.hover_fact:
            # match either content or id
            if (self.hover_fact == self.current_fact
                    or (self.hover_fact
                        and self.current_fact
                        and self.hover_fact.id == self.current_fact.id)
                    ):
                self.unset_current_fact()
            # Totals can't be selected
            elif not isinstance(self.hover_fact, TotalFact):
                self.set_current_fact(self.hover_fact)

    def activate_row(self, day, fact):
        self.emit("on-activate-row", day, fact)

    def delete_row(self, fact):
        self.emit("on-delete-called", fact)

    def on_double_click(self, scene, event):
        if self.hover_fact and not isinstance(self.hover_fact, TotalFact):
            self.activate_row(self.hover_day, self.hover_fact)

    def on_key_press(self, scene, event):
        # all keys should appear also in the Overview.on_key_press
        # to be forwarded here even without focus.
        if event.keyval == gdk.KEY_Up:
            if self.facts:
                if self.current_fact:
                    idx = max(0, self.current_fact_index - 1)
                else:
                    # enter from below
                    idx = len(self.facts) - 1
                self.set_current_fact(self.facts[idx])

        elif event.keyval == gdk.KEY_Down:
            if self.facts:
                if self.current_fact:
                    idx = min(len(self.facts) - 1, self.current_fact_index + 1)
                else:
                    # enter from top
                    idx = 0
                self.set_current_fact(self.facts[idx])

        elif event.keyval == gdk.KEY_Home:
            if self.facts:
                self.set_current_fact(self.facts[0])

        elif event.keyval == gdk.KEY_End:
            if self.facts:
                self.set_current_fact(self.facts[-1])

        elif event.keyval == gdk.KEY_Page_Down:
            self.y += self.height * 0.8
            self.on_scroll()

        elif event.keyval == gdk.KEY_Page_Up:
            self.y -= self.height * 0.8
            self.on_scroll()

        elif event.keyval == gdk.KEY_Return:
            if self.current_fact:
                self.activate_row(self.hover_day, self.current_fact)

        elif event.keyval == gdk.KEY_Delete:
            if self.current_fact:
                self.delete_row(self.current_fact)

    def set_current_fact(self, fact):
        self.current_fact = fact

        if fact.y < self.y:
            self.y = fact.y
        if (fact.y + fact.height) > (self.y + self.height):
            self.y = fact.y + fact.height - self.height

        self.on_scroll()

    def unset_current_fact(self):
        """Deselect fact."""
        self.current_fact = None
        self.on_scroll()

    def get_visible_range(self):
        start, end = (bisect.bisect(self.row_positions, self.y) - 1,
                      bisect.bisect(self.row_positions, self.y + self.height))

        y = self.y
        return [{"i": start + i, "y": pos - y, "h": height, "day": day, "facts": facts}
                for i, (pos, height, (day, facts)) in enumerate(zip(self.row_positions[start:end],
                                                                    self.row_heights[start:end],
                                                                    self.days[start:end]))]

    def on_mouse_move(self, tree, event):
        hover_day, hover_fact = None, None

        for rec in self.visible_range:
            if rec['y'] <= event.y <= (rec['y'] + rec['h']):
                hover_day = rec
                break

        if hover_day != self.hover_day:
            # Facts are considered equal if their content is the same,
            # even if their id is different.
            # redraw only cares about content, not id.
            self.redraw()
        # make sure it is always fully updated, including facts ids.
        self.hover_day = hover_day

        if self.hover_day:
            for fact in self.hover_day.get('facts', []):
                if (fact.y - self.y) <= event.y <= (fact.y - self.y + fact.height):
                    hover_fact = fact
                    break

        if (hover_fact
                and self.hover_fact
                and hover_fact.id != self.hover_fact.id
                ):
            self.move_actions()
        # idem, always update hover_fact, not just if they appear different
        self.hover_fact = hover_fact

    def move_actions(self):
        if self.hover_fact:
            self.action_row.visible = True
            self.action_row.x = self.width - 80 - self.action_row.width
            self.action_row.y = self.hover_fact.y - self.y
        else:
            self.action_row.visible = False

    def _on_vadjustment_change(self, scene, vadjustment):
        if not self.vadjustment:
            return
        self.vadjustment.connect("value_changed", self.on_scroll_value_changed)
        self.set_size_request(500, 300)

    def set_facts(self, facts, scroll_to_top=False):
        # FactTree adds attributes to its facts. isolate these side effects
        # copy the id too; most of the checks are based on id here.
        self.facts = [fact.copy(id=fact.id) for fact in facts]
        del facts  # make sure facts is not used by inadvertance below.

        # If we get an entirely new set of facts, scroll back to the top
        if scroll_to_top:
            self.y = 0
        self.hover_fact = None
        if self.vadjustment:
            self.vadjustment.set_value(self.y)

        if self.facts:
            start = self.facts[0].date
            end = self.facts[-1].date
        else:
            start = end = dt.hday.today()

        by_date = defaultdict(list)
        delta_by_date = defaultdict(dt.timedelta)
        for fact in self.facts:
            by_date[fact.date].append(fact)
            delta_by_date[fact.date] += fact.delta

        # Add a TotalFact at the end of each day if we are
        # displaying more than one day.
        if len(by_date) > 1:
            for key in by_date:
                total_by_date = TotalFact(_("Total"), delta_by_date[key])
                by_date[key].append(total_by_date)

        days = []
        for i in range((end - start).days + 1):
            current_date = start + dt.timedelta(days=i)
            if current_date in by_date:
                days.append((current_date, by_date[current_date]))

        self.days = days

        self.set_row_heights()

        if (self.current_fact
                and self.current_fact.id in (fact.id for fact in self.facts)
                ):
            self.on_scroll()
        else:
            # will also trigger an on_scroll
            self.unset_current_fact()

    def set_row_heights(self):
        """
            the row height is defined by following factors:
                * how many facts are there in the day
                * does the fact have description / tags

            This func creates a list of row start positions to be able to
            quickly determine what to display
        """
        if not self.height:
            return

        y, pos, heights = 0, [], []

        for date, facts in self.days:
            height = 0
            for fact in facts:
                self.fact_row.set_fact(fact)
                fact_height = self.fact_row.height
                fact.y = y + height
                fact.height = fact_height

                height += fact.height

            height += self.day_padding
            height = max(height, 60)

            pos.append(y)
            heights.append(height)
            y += height

        self.row_positions, self.row_heights = pos, heights

        maxy = max(y, 1)

        if self.vadjustment:
            self.vadjustment.set_lower(0)
            self.vadjustment.set_upper(max(maxy, self.height))
            self.vadjustment.set_page_size(self.height)

    def on_resize(self, scene, event):
        self.set_row_heights()
        self.fact_row.width = self.width - 105
        self.on_scroll()

    def on_scroll_value_changed(self, scroll):
        self.y = int(scroll.get_value())
        self.on_scroll()

    def on_scroll(self, scene=None, event=None):
        if not self.height:
            return
        y_pos = self.y
        direction = 0
        if event and event.direction == gdk.ScrollDirection.UP:
            direction = -1
        elif event and event.direction == gdk.ScrollDirection.DOWN:
            direction = 1

        y_pos += 15 * direction
        if self.vadjustment:
            y_pos = max(0, min(self.vadjustment.get_upper() - self.height, y_pos))
            self.vadjustment.set_value(y_pos)
        self.y = y_pos

        self.move_actions()
        self.redraw()

        self.visible_range = self.get_visible_range()

    def on_enter_frame(self, scene, context):
        has_focus = self.get_toplevel().has_toplevel_focus()
        if has_focus:
            colors = {
                "normal": self.style.get_color(gtk.StateFlags.NORMAL),
                "normal_bg": self.style.get_background_color(gtk.StateFlags.NORMAL),
                "selected": self.style.get_color(gtk.StateFlags.SELECTED),
                "selected_bg": self.style.get_background_color(gtk.StateFlags.SELECTED),
            }
        else:
            colors = {
                "normal": self.style.get_color(gtk.StateFlags.BACKDROP),
                "normal_bg": self.style.get_background_color(gtk.StateFlags.BACKDROP),
                "selected": self.style.get_color(gtk.StateFlags.BACKDROP),
                "selected_bg": self.style.get_background_color(gtk.StateFlags.BACKDROP),
            }

        if not self.height:
            return

        g = graphics.Graphics(context)

        g.set_line_style(1)
        g.translate(0.5, 0.5)

        date_bg_color = self.colors.mix(colors["normal_bg"], colors["normal"], 0.15)
        g.fill_area(0, 0, 105, self.height, date_bg_color)

        y = int(self.y)

        for rec in self.visible_range:
            g.save_context()
            g.translate(0, rec['y'])
            g.set_color(colors["normal"])
            self.date_label.show(g, rec['day'].strftime("%A\n%b %d"))

            g.translate(105, 0)
            for fact in rec['facts']:
                is_selected = (self.current_fact is not None
                               and fact.id == self.current_fact.id)
                self.fact_row.set_fact(fact)
                self.fact_row.show(g, colors, is_selected=is_selected)
                g.translate(0, self.fact_row.height)

            g.restore_context()



# This file is part of Hamster
# Copyright (c) The Hamster time tracker developers
# SPDX-License-Identifier: GPL-3.0-or-later


"""Fact definition."""


import logging
logger = logging.getLogger(__name__)   # noqa: E402

import calendar

from copy import deepcopy

from hamster.lib import datetime as dt
from hamster.lib.parsing import parse_fact, get_tags_from_description


class FactError(Exception):
    """Generic Fact error."""


class Fact(object):
    def __init__(self, activity="", category=None, description=None, tags=None,
                 range=None, start=None, end=None, start_time=None, end_time=None,
                 id=None, activity_id=None):
        """Homogeneous chunk of activity.

        The category, description and tags must be passed explicitly.

        To provide the whole fact information as a single string,
        please use Fact.parse(string).

        range (dt.Range): time spanned by the fact. For convenience,
                          the `start` and `end` arguments can be given instead.
        start (dt.datetime); Start of the fact range.
                             Mutually exclusive with `range`.
        end (dt.datetime); End of the fact range.
                           Mutually exclusive with `range`.
        start_time (dt.datetime): Deprecated. Same as start.
        end_time (dt.datetime): Deprecated. Same as end.

        id (int): id in the database.
                  Should be used with extreme caution, knowing exactly why.
                  (only for very specific direct database read/write)
        """

        self.activity = activity
        self.category = category
        self.description = description
        self.tags = tags or []
        if range:
            assert not start, "range already given"
            assert not end, "range already given"
            assert not start_time, "range already given"
            assert not end_time, "range already given"
            self.range = range
        else:
            if start_time:
                assert not start, "use only start, not start_time"
                start = start_time
            if end_time:
                assert not end, "use only end, not end_time"
                end = end_time
            self.range = dt.Range(start, end)
        self.id = id
        self.activity_id = activity_id

    # TODO: might need some cleanup
    def as_dict(self):
        date = self.date
        return {
            'id': int(self.id) if self.id else "",
            'activity': self.activity,
            'category': self.category,
            'description': self.description,
            'tags': [tag.strip() for tag in self.tags],
            'date': calendar.timegm(date.timetuple()) if date else "",
            'start_time': self.range.start if isinstance(self.range.start, str) else calendar.timegm(self.range.start.timetuple()),
            'end_time': self.range.end if isinstance(self.range.end, str) else calendar.timegm(self.range.end.timetuple()) if self.range.end else "",
            'delta': self.delta.total_seconds()  # ugly, but needed for report.py
        }

    @property
    def activity(self):
        """Activity name."""
        return self._activity

    @activity.setter
    def activity(self, value):
        self._activity = value.strip() if value else ""

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, value):
        self._category = value.strip() if value else ""

    def copy(self, **kwds):
        """Return an independent copy, with overrides as keyword arguments.

        By default, only copy user-visible attributes.
        To also copy the id, use fact.copy(id=fact.id)
        """
        fact = deepcopy(self)
        fact._set(**kwds)
        return fact

    @property
    def date(self):
        """hamster day, determined from start_time.

        Note: Setting date is a one-shot modification of
              the start_time and end_time (if defined),
              to match the given value.
              Any subsequent modification of start_time
              can result in different self.date.
        """
        return self.range.start.hday() if self.range.start else None

    @date.setter
    def date(self, value):
        if self.range.start:
            previous_start_time = self.range.start
            self.range.start = dt.datetime.from_day_time(value, self.range.start.time())
            if self.range.end:
                # start_time date prevails.
                # Shift end_time to preserve the fact duration.
                self.range.end += self.range.start - previous_start_time
        elif self.range.end:
            self.range.end = dt.datetime.from_day_time(value, self.range.end.time())

    @property
    def delta(self):
        """Duration (datetime.timedelta)."""
        end_time = self.range.end if self.range.end else dt.datetime.now()
        return end_time - self.range.start

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value.strip() if value else ""

    @property
    def end_time(self):
        """Fact range end.

        Deprecated, use self.range.end instead.
        """
        return self.range.end

    @end_time.setter
    def end_time(self, value):
        self.range.end = value

    @property
    def start_time(self):
        """Fact range start.

        Deprecated, use self.range.start instead.
        """
        return self.range.start

    @start_time.setter
    def start_time(self, value):
        self.range.start = value

    @classmethod
    def parse(cls, string, range_pos="head", default_day=None, ref="now"):
        fact = Fact()
        for key, val in parse_fact(string, range_pos=range_pos,
                                   default_day=default_day, ref=ref).items():
            setattr(fact, key, val)
        return fact

    def serialized_name(self):
        res = self.activity

        if self.category:
            res += "@%s" % self.category

        if self.description:
            res += ', '
            res += self.description

        if self.tags:
            # Don't duplicate tags that are already in the description
            seen_tags = get_tags_from_description(self.description)
            remaining_tags = [
                tag for tag in self.tags if tag not in seen_tags
            ]
            if remaining_tags:
                res += ", %s" % " ".join("#%s" % tag for tag in remaining_tags)
        return res

    def serialized(self, range_pos="head", default_day=None):
        """Return a string fully representing the fact."""
        name = self.serialized_name()
        if range_pos == "head":
            # Is activity starting range-like ?
            subfact = Fact.parse(self.activity, range_pos=range_pos,
                                 default_day=default_day)
            need_explicit = bool(subfact.range)
        else:
            # TODO: should check last tag.
            need_explicit = False
        datetime = self.range.format(default_day=default_day,
                                     explicit_none=need_explicit)
        # no need for space if name or datetime is missing
        space = " " if name and datetime else ""
        assert range_pos in ("head", "tail")
        if range_pos == "head":
            return "{}{}{}".format(datetime, space, name)
        else:
            return "{}{}{}".format(name, space, datetime)

    def _set(self, **kwds):
        """Modify attributes.

        Private, used only in copy. It is more readable to be explicit, e.g.:
        fact.range.start = ...
        fact.range.end = ...
        """
        for attr, value in kwds.items():
            if not hasattr(self, attr):
                raise AttributeError("'{attr}' not found".format(attr=attr))
            else:
                setattr(self, attr, value)

    def __eq__(self, other):
        return (isinstance(other, Fact)
                and self.activity == other.activity
                and self.category == other.category
                and self.description == other.description
                and self.range.end == other.range.end
                and self.range.start == other.range.start
                and self.tags == other.tags
                )

    def __repr__(self):
        return self.serialized(default_day=None)



# -*- coding: utf-8 -*-

# Copyright (C) 2007-2010 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

import time

from gi.repository import Gtk as gtk
from gi.repository import GObject as gobject
from gi.repository import PangoCairo as pangocairo

from hamster.lib import datetime as dt
from hamster.lib import graphics, pytweener
from hamster.lib.configuration import conf


class Selection(graphics.Sprite):
    def __init__(self, start_time = None, end_time = None):
        graphics.Sprite.__init__(self, z_order = 100)
        self.start_time, self.end_time  = None, None
        self.width, self.height = None, None
        self.fill = None # will be set to proper theme color on render
        self.fixed = False

        self.start_label = graphics.Label("", 11, "#333", visible = False)
        self.end_label = graphics.Label("", 11, "#333", visible = False)
        self.duration_label = graphics.Label("", 11, "#FFF", visible = False)

        self.add_child(self.start_label, self.end_label, self.duration_label)
        self.connect("on-render", self.on_render)


    def on_render(self, sprite):
        if not self.fill: # not ready yet
            return

        self.graphics.rectangle(0, 0, self.width, self.height)
        self.graphics.fill_preserve(self.fill, 0.3)
        self.graphics.stroke(self.fill)


        # adjust labels
        self.start_label.visible = self.start_time is not None and self.start_time != self.end_time
        if self.start_label.visible:
            self.start_label.text = self.start_time.strftime("%H:%M")
            if self.x - self.start_label.width - 5 > 0:
                self.start_label.x = -self.start_label.width - 5
            else:
                self.start_label.x = 5

            self.start_label.y = self.height + 2

        self.end_label.visible = self.end_time is not None and self.start_time != self.end_time
        if self.end_label.visible:
            self.end_label.text = self.end_time.strftime("%H:%M")
            self.end_label.x = self.width + 5
            self.end_label.y = self.height + 2



            duration = self.end_time - self.start_time
            duration = int(duration.seconds / 60)
            self.duration_label.text =  "%02d:%02d" % (duration / 60, duration % 60)

            self.duration_label.visible = self.duration_label.width < self.width
            if self.duration_label.visible:
                self.duration_label.y = (self.height - self.duration_label.height) / 2
                self.duration_label.x = (self.width - self.duration_label.width) / 2
        else:
            self.duration_label.visible = False



class DayLine(graphics.Scene):
    def __init__(self, start_time = None):
        graphics.Scene.__init__(self)
        self.set_can_focus(False) # no interaction

        self.day_start = conf.day_start

        start_time = start_time or dt.datetime.now()

        self.view_time = start_time or dt.datetime.combine(start_time.date(), self.day_start)

        self.scope_hours = 24


        self.fact_bars = []
        self.categories = []

        self.connect("on-enter-frame", self.on_enter_frame)

        self.plot_area = graphics.Sprite(y=15)

        self.chosen_selection = Selection()
        self.plot_area.add_child(self.chosen_selection)

        self.drag_start = None
        self.current_x = None

        self.date_label = graphics.Label(color=self._style.get_color(gtk.StateFlags.NORMAL),
                                         x=5, y=16)

        self.add_child(self.plot_area, self.date_label)


    def plot(self, date, facts, select_start, select_end = None):
        for bar in self.fact_bars:
            self.plot_area.sprites.remove(bar)

        self.fact_bars = []
        for fact in facts:
            fact_bar = graphics.Rectangle(0, 0, fill="#aaa", stroke="#aaa") # dimensions will depend on screen situation
            fact_bar.fact = fact

            if fact.category in self.categories:
                fact_bar.category = self.categories.index(fact.category)
            else:
                fact_bar.category = len(self.categories)
                self.categories.append(fact.category)

            self.plot_area.add_child(fact_bar)
            self.fact_bars.append(fact_bar)

        self.view_time = dt.datetime.combine(date, self.day_start)
        self.date_label.text = self.view_time.strftime("%b %d")

        self.chosen_selection.start_time = select_start
        self.chosen_selection.end_time = select_end

        self.chosen_selection.width = None
        self.chosen_selection.fixed = True
        self.chosen_selection.visible = True

        self.redraw()


    def on_enter_frame(self, scene, context):
        g = graphics.Graphics(context)

        self.plot_area.height = self.height - 30


        vertical = min(self.plot_area.height / 5, 7)
        minute_pixel = (self.scope_hours * 60.0 - 15) / self.width

        g.set_line_style(width=1)
        g.translate(0.5, 0.5)


        colors = {
            "normal": self._style.get_color(gtk.StateFlags.NORMAL),
            "normal_bg": self._style.get_background_color(gtk.StateFlags.NORMAL),
            "selected": self._style.get_color(gtk.StateFlags.SELECTED),
            "selected_bg": self._style.get_background_color(gtk.StateFlags.SELECTED),
        }

        bottom = self.plot_area.y + self.plot_area.height

        for bar in self.fact_bars:
            bar.y = vertical * bar.category + 5
            bar.height = vertical

            bar_start_time = bar.fact.start_time - self.view_time
            minutes = bar_start_time.seconds / 60 + bar_start_time.days * self.scope_hours  * 60

            bar.x = round(minutes / minute_pixel) + 0.5
            bar.width = round((bar.fact.delta).seconds / 60 / minute_pixel)


        if self.chosen_selection.start_time and self.chosen_selection.width is None:
            # we have time but no pixels
            minutes = round((self.chosen_selection.start_time - self.view_time).seconds / 60 / minute_pixel) + 0.5
            self.chosen_selection.x = minutes
            if self.chosen_selection.end_time:
                self.chosen_selection.width = round((self.chosen_selection.end_time - self.chosen_selection.start_time).seconds / 60 / minute_pixel)
            else:
                self.chosen_selection.width = 0
            self.chosen_selection.height = self.chosen_selection.parent.height

            # use the oportunity to set proper colors too
            self.chosen_selection.fill = colors['selected_bg']
            self.chosen_selection.duration_label.color = colors['selected']




        #time scale
        g.set_color("#000")

        background = colors["normal_bg"]
        text = colors["normal"]

        tick_color = g.colors.contrast(background, 80)

        layout = g.create_layout(size = 10)
        for i in range(self.scope_hours * 60):
            time = (self.view_time + dt.timedelta(minutes=i))

            g.set_color(tick_color)
            if time.minute == 0:
                g.move_to(round(i / minute_pixel), bottom - 15)
                g.line_to(round(i / minute_pixel), bottom)
                g.stroke()
            elif time.minute % 15 == 0:
                g.move_to(round(i / minute_pixel), bottom - 5)
                g.line_to(round(i / minute_pixel), bottom)
                g.stroke()



            if time.minute == 0 and time.hour % 4 == 0:
                if time.hour == 0:
                    g.move_to(round(i / minute_pixel), self.plot_area.y)
                    g.line_to(round(i / minute_pixel), bottom)
                    label_minutes = time.strftime("%b %d")
                else:
                    label_minutes = time.strftime("%H<small><sup>%M</sup></small>")

                g.set_color(text)
                layout.set_markup(label_minutes)

                g.move_to(round(i / minute_pixel) + 2, 0)
                pangocairo.show_layout(context, layout)

        #current time
        if self.view_time < dt.datetime.now() < self.view_time + dt.timedelta(hours = self.scope_hours):
            minutes = round((dt.datetime.now() - self.view_time).seconds / 60 / minute_pixel)
            g.rectangle(minutes, 0, self.width, self.height)
            g.fill(colors['normal_bg'], 0.7)

            g.move_to(minutes, self.plot_area.y)
            g.line_to(minutes, bottom)
            g.stroke("#f00", 0.4)



import dbus

from dbus.mainloop.glib import DBusGMainLoop as DBusMainLoop
from json import dumps, loads
from calendar import timegm
from hamster.lib import datetime as dt
from hamster.lib.fact import Fact


"""D-Bus communication utilities."""

# file layout: functions sorted in alphabetical order,
# not taking into account the "to_" and "from_" prefixes.
# So back and forth conversions are close to one another.


# dates

def from_dbus_date(dbus_date):
    """Convert D-Bus timestamp (seconds since epoch) to date."""
    return dt.date.fromtimestamp(dbus_date) if dbus_date else None


def to_dbus_date(date):
    """Convert date to D-Bus timestamp (seconds since epoch)."""
    return timegm(date.timetuple()) if date else 0


# facts

def from_dbus_fact_json(dbus_fact):
    """Convert D-Bus JSON to Fact."""
    d = loads(dbus_fact)
    range_d = d['range']
    # should use pdt.datetime.fromisoformat,
    # but that appears only in python3.7, nevermind
    start_s = range_d['start']
    end_s = range_d['end']
    range = dt.Range(start=dt.datetime.parse(start_s) if start_s else None,
                     end=dt.datetime.parse(end_s) if end_s else None)
    d['range'] = range
    return Fact(**d)


def to_dbus_fact_json(fact):
    """Convert Fact to D-Bus JSON (str)."""
    d = {}
    keys = ('activity', 'category', 'description', 'tags', 'id', 'activity_id')
    for key in keys:
        d[key] = getattr(fact, key)
    # isoformat(timespec="minutes") appears only in python3.6, nevermind
    # and fromisoformat is not available anyway, so let's talk hamster
    start = str(fact.range.start) if fact.range.start else None
    end = str(fact.range.end) if fact.range.end else None
    d['range'] = {'start': start, 'end': end}
    return dumps(d)


# Range

def from_dbus_range(dbus_range):
    """Convert from D-Bus string to dt.Range."""
    range, __ = dt.Range.parse(dbus_range, position="exact")
    return range


def to_dbus_range(range):
    """Convert dt.Range to D-Bus string."""

    # no default_day, to always output in the same format
    return range.format(default_day=None)


# Legacy functions:

"""
old dbus_fact signature (types matching the to_dbus_fact output)
    i  id
    i  start_time
    i  end_time
    s  description
    s  activity name
    i  activity id
    s  category name
    as List of fact tags
    i  date
    i  delta
"""
fact_signature = '(iiissisasii)'


def from_dbus_fact(dbus_fact):
    """Unpack the struct into a proper dict.

    Legacy: to besuperceded by from_dbus_fact_json at some point.
    """
    return Fact(activity=dbus_fact[4],
                start_time=dt.datetime.utcfromtimestamp(dbus_fact[1]),
                end_time=dt.datetime.utcfromtimestamp(dbus_fact[2]) if dbus_fact[2] else None,
                description=dbus_fact[3],
                activity_id=dbus_fact[5],
                category=dbus_fact[6],
                tags=dbus_fact[7],
                id=dbus_fact[0]
                )


def to_dbus_fact(fact):
    """Perform Fact conversion to D-Bus.

    Return the corresponding dbus structure, with supported data types.
    Legacy: to besuperceded by to_dbus_fact_json at some point.
    """
    return (fact.id or 0,
            timegm(fact.start_time.timetuple()),
            timegm(fact.end_time.timetuple()) if fact.end_time else 0,
            fact.description or '',
            fact.activity or '',
            fact.activity_id or 0,
            fact.category or '',
            dbus.Array(fact.tags, signature = 's'),
            to_dbus_date(fact.date),
            fact.delta.days * 24 * 60 * 60 + fact.delta.seconds)



# -*- coding: utf-8 -*-

# Copyright (C) 2007, 2008, 2014 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import GObject as gobject

from hamster import widgets
from hamster.lib import datetime as dt
from hamster.lib import stuff
from hamster.lib.configuration import Controller, runtime, conf


def get_prev(selection, model):
    (model, iter) = selection.get_selected()

    #previous item
    path = model.get_path(iter)[0] - 1
    if path >= 0:
        return model.get_iter_from_string(str(path))
    else:
        return None


class CategoryStore(gtk.ListStore):
    def __init__(self):
        #id, name, color_code, order
        gtk.ListStore.__init__(self, int, str)

    def load(self):
        category_list = runtime.storage.get_categories()

        for category in category_list:
            self.append([category['id'], category['name']])

        self.unsorted_category = self.append([-1, _("Unsorted")]) # all activities without category


class ActivityStore(gtk.ListStore):
    def __init__(self):
        #id, name, category_id, order
        gtk.ListStore.__init__(self, int, str, int)

    def load(self, category_id):
        self.clear()

        if category_id is None:
            return

        activity_list = runtime.storage.get_category_activities(category_id)

        for activity in activity_list:
            self.append([activity['id'],
                         activity['name'],
                         activity['category_id']])


class PreferencesEditor(Controller):
    TARGETS = [
        ('MY_TREE_MODEL_ROW', gtk.TargetFlags.SAME_WIDGET, 0),
        ('MY_TREE_MODEL_ROW', gtk.TargetFlags.SAME_APP, 0),
        ]

    def __init__(self):
        Controller.__init__(self, ui_file="preferences.ui")

        # create and fill activity tree
        self.activity_tree = self.get_widget('activity_list')
        self.get_widget("activities_label").set_mnemonic_widget(self.activity_tree)
        self.activity_store = ActivityStore()

        self.external_listeners = []

        self.activityColumn = gtk.TreeViewColumn(_("Name"))
        self.activityColumn.set_expand(True)
        self.activityCell = gtk.CellRendererText()
        self.external_listeners.extend([
            (self.activityCell, self.activityCell.connect('edited', self.activity_name_edited_cb, self.activity_store))
        ])
        self.activityColumn.pack_start(self.activityCell, True)
        self.activityColumn.set_attributes(self.activityCell, text=1)
        self.activityColumn.set_sort_column_id(1)
        self.activity_tree.append_column(self.activityColumn)

        self.activity_tree.set_model(self.activity_store)

        self.selection = self.activity_tree.get_selection()

        self.external_listeners.extend([
            (self.selection, self.selection.connect('changed', self.activity_changed, self.activity_store))
        ])

        # create and fill category tree
        self.category_tree = self.get_widget('category_list')
        self.get_widget("categories_label").set_mnemonic_widget(self.category_tree)
        self.category_store = CategoryStore()

        self.categoryColumn = gtk.TreeViewColumn(_("Category"))
        self.categoryColumn.set_expand(True)
        self.categoryCell = gtk.CellRendererText()

        self.external_listeners.extend([
            (self.categoryCell, self.categoryCell.connect('edited', self.category_edited_cb, self.category_store))
        ])

        self.categoryColumn.pack_start(self.categoryCell, True)
        self.categoryColumn.set_attributes(self.categoryCell, text=1)
        self.categoryColumn.set_sort_column_id(1)
        self.categoryColumn.set_cell_data_func(self.categoryCell, self.unsorted_painter)
        self.category_tree.append_column(self.categoryColumn)

        self.category_store.load()
        self.category_tree.set_model(self.category_store)

        selection = self.category_tree.get_selection()
        self.external_listeners.extend([
            (selection, selection.connect('changed', self.category_changed_cb, self.category_store))
        ])

        self.day_start = widgets.TimeInput(dt.time(5,30), parent=self.get_widget("day_start_placeholder"))

        self.load_config()

        # Allow enable drag and drop of rows including row move
        self.activity_tree.enable_model_drag_source(gdk.ModifierType.BUTTON1_MASK,
                                                    self.TARGETS,
                                                    gdk.DragAction.DEFAULT|
                                                    gdk.DragAction.MOVE)

        self.category_tree.enable_model_drag_dest(self.TARGETS,
                                                  gdk.DragAction.MOVE)

        self.activity_tree.connect("drag_data_get", self.drag_data_get_data)

        self.category_tree.connect("drag_data_received", self.on_category_drop)

        #select first category
        selection = self.category_tree.get_selection()
        selection.select_path((0,))

        self.prev_selected_activity = None
        self.prev_selected_category = None

        self.external_listeners.extend([
            (self.day_start, self.day_start.connect("time-entered", self.on_day_start_changed))
        ])

        self.show()

    def show(self):
        self.get_widget("notebook1").set_current_page(0)
        self.window.show_all()

    def load_config(self, *args):
        self.day_start.time = conf.day_start

        self.tags = [tag["name"] for tag in runtime.storage.get_tags(only_autocomplete=True)]
        self.get_widget("autocomplete_tags").set_text(", ".join(self.tags))

    def on_autocomplete_tags_view_focus_out_event(self, view, event):
        buf = self.get_widget("autocomplete_tags")
        updated_tags = buf.get_text(buf.get_start_iter(), buf.get_end_iter(), 0)
        if updated_tags == self.tags:
            return

        self.tags = updated_tags

        runtime.storage.update_autocomplete_tags(updated_tags)

    def drag_data_get_data(self, treeview, context, selection, target_id,
                           etime):
        treeselection = treeview.get_selection()
        model, iter = treeselection.get_selected()
        data = model.get_value(iter, 0) #get activity ID
        selection.set(selection.target, 0, str(data))

    def select_activity(self, id):
        model = self.activity_tree.get_model()
        i = 0
        for row in model:
            if row[0] == id:
                self.activity_tree.set_cursor((i, ))
            i += 1

    def select_category(self, id):
        model = self.category_tree.get_model()
        i = 0
        for row in model:
            if row[0] == id:
                self.category_tree.set_cursor((i, ))
            i += 1

    def on_category_list_drag_motion(self, treeview, drag_context, x, y, eventtime):
        self.prev_selected_category = None
        try:
            target_path, drop_position = treeview.get_dest_row_at_pos(x, y)
            model, source = treeview.get_selection().get_selected()

        except:
            return

        drop_yes = ("drop_yes", gtk.TargetFlags.SAME_APP, 0)
        drop_no = ("drop_no", gtk.TargetFlags.SAME_APP, 0)

        if drop_position != gtk.TREE_VIEW_DROP_AFTER and \
           drop_position != gtk.TREE_VIEW_DROP_BEFORE:
            treeview.enable_model_drag_dest(self.TARGETS, gdk.DragAction.MOVE)
        else:
            treeview.enable_model_drag_dest([drop_no], gdk.DragAction.MOVE)

    def on_category_drop(self, treeview, context, x, y, selection,
                                info, etime):
        model = self.category_tree.get_model()
        data = selection.data
        drop_info = treeview.get_dest_row_at_pos(x, y)

        if drop_info:
            path, position = drop_info
            iter = model.get_iter(path)
            changed = runtime.storage.change_category(int(data), model[iter][0])

            context.finish(changed, True, etime)
        else:
            context.finish(False, True, etime)

        return

    # callbacks
    def category_edited_cb(self, cell, path, new_text, model):
        id = model[path][0]
        if id == -1:
            return False #ignoring unsorted category

        #look for dupes
        categories = runtime.storage.get_categories()
        for category in categories:
            if category['name'].lower() == new_text.lower():
                if id == -2: # that was a new category
                    self.category_store.remove(model.get_iter(path))
                self.select_category(category['id'])
                return False

        if id == -2: #new category
            id = runtime.storage.add_category(new_text)
            model[path][0] = id
        else:
            runtime.storage.update_category(id, new_text)

        model[path][1] = new_text

    def activity_name_edited_cb(self, cell, path, new_text, model):
        id = model[path][0]
        category_id = model[path][2]

        activities = runtime.storage.get_category_activities(category_id)
        prev = None
        for activity in activities:
            if id == activity['id']:
                prev = activity['name']
            else:
                # avoid two activities in same category with same name
                if activity['name'].lower() == new_text.lower():
                    if id == -1: # that was a new activity
                        self.activity_store.remove(model.get_iter(path))
                    self.select_activity(activity['id'])
                    return False

        if id == -1: #new activity -> add
            model[path][0] = runtime.storage.add_activity(new_text, category_id)
        else: #existing activity -> update
            new = new_text
            runtime.storage.update_activity(id, new, category_id)

        model[path][1] = new_text
        return True

    def category_changed_cb(self, selection, model):
        """ enables and disables action buttons depending on selected item """
        (model, iter) = selection.get_selected()
        id = 0
        if iter is None:
            self.activity_store.clear()
        else:
            self.prev_selected_activity = None

            id = model[iter][0]
            self.activity_store.load(model[iter][0])

        #start with nothing
        self.get_widget('activity_edit').set_sensitive(False)
        self.get_widget('activity_remove').set_sensitive(False)

        return True

    def _get_selected_category(self):
        selection = self.get_widget('category_list').get_selection()
        (model, iter) = selection.get_selected()

        return model[iter][0] if iter else None

    def activity_changed(self, selection, model):
        """ enables and disables action buttons depending on selected item """
        (model, iter) = selection.get_selected()

        # treat any selected case
        unsorted_selected = self._get_selected_category() == -1
        self.get_widget('activity_edit').set_sensitive(iter != None)
        self.get_widget('activity_remove').set_sensitive(iter != None)

    def _del_selected_row(self, tree):
        selection = tree.get_selection()
        (model, iter) = selection.get_selected()

        next_row = model.iter_next(iter)

        if next_row:
            selection.select_iter(next_row)
        else:
            path = model.get_path(iter)[0] - 1
            if path > 0:
                selection.select_path(path)

        removable_id = model[iter][0]
        model.remove(iter)
        return removable_id

    def unsorted_painter(self, column, cell, model, iter, data):
        cell_id = model.get_value(iter, 0)
        cell_text = model.get_value(iter, 1)
        if cell_id == -1:
            text = '<span color="#555" style="italic">%s</span>' % cell_text # TODO - should get color from theme
            cell.set_property('markup', text)
        else:
            cell.set_property('text', cell_text)

        return

    def on_activity_list_button_pressed(self, tree, event):
        self.activityCell.set_property("editable", False)

    def on_activity_list_button_released(self, tree, event):
        if event.button == 1 and tree.get_path_at_pos(int(event.x), int(event.y)):
            # Get treeview path.
            path, column, x, y = tree.get_path_at_pos(int(event.x), int(event.y))

            if self.prev_selected_activity == path:
                self.activityCell.set_property("editable", True)
                tree.set_cursor_on_cell(path, self.activityColumn, self.activityCell, True)

            self.prev_selected_activity = path

    def on_category_list_button_pressed(self, tree, event):
        self.activityCell.set_property("editable", False)

    def on_category_list_button_released(self, tree, event):
        if event.button == 1 and tree.get_path_at_pos(int(event.x), int(event.y)):
            # Get treeview path.
            path, column, x, y = tree.get_path_at_pos(int(event.x), int(event.y))

            if self.prev_selected_category == path and \
               self._get_selected_category() != -1: #do not allow to edit unsorted
                self.categoryCell.set_property("editable", True)
                tree.set_cursor_on_cell(path, self.categoryColumn, self.categoryCell, True)
            else:
                self.categoryCell.set_property("editable", False)


            self.prev_selected_category = path

    def on_activity_remove_clicked(self, button):
        self.remove_current_activity()

    def on_activity_edit_clicked(self, button):
        self.activityCell.set_property("editable", True)

        selection = self.activity_tree.get_selection()
        (model, iter) = selection.get_selected()
        path = model.get_path(iter)
        self.activity_tree.set_cursor_on_cell(path, self.activityColumn, self.activityCell, True)

    """keyboard events"""
    def on_activity_list_key_pressed(self, tree, event_key):
        key = event_key.keyval
        selection = tree.get_selection()
        (model, iter) = selection.get_selected()
        if (event_key.keyval == gdk.KEY_Delete):
            self.remove_current_activity()

        elif key == gdk.KEY_F2 :
            self.activityCell.set_property("editable", True)
            path = model.get_path(iter)
            tree.set_cursor_on_cell(path, self.activityColumn, self.activityCell, True)

    def remove_current_activity(self):
        selection = self.activity_tree.get_selection()
        (model, iter) = selection.get_selected()
        runtime.storage.remove_activity(model[iter][0])
        self._del_selected_row(self.activity_tree)

    def on_category_remove_clicked(self, button):
        self.remove_current_category()

    def on_category_edit_clicked(self, button):
        self.categoryCell.set_property("editable", True)

        selection = self.category_tree.get_selection()
        (model, iter) = selection.get_selected()
        path = model.get_path(iter)
        self.category_tree.set_cursor_on_cell(path, self.categoryColumn, self.categoryCell, True)

    def on_category_list_key_pressed(self, tree, event_key):
        key = event_key.keyval

        if self._get_selected_category() == -1:
            return #ignoring unsorted category

        selection = tree.get_selection()
        (model, iter) = selection.get_selected()

        if  key == gdk.KEY_Delete:
            self.remove_current_category()
        elif key == gdk.KEY_F2:
            self.categoryCell.set_property("editable", True)
            path = model.get_path(iter)
            tree.set_cursor_on_cell(path, self.categoryColumn, self.categoryCell, True)

    def remove_current_category(self):
        selection = self.category_tree.get_selection()
        (model, iter) = selection.get_selected()
        id = model[iter][0]
        if id != -1:
            runtime.storage.remove_category(id)
            self._del_selected_row(self.category_tree)

    def on_preferences_window_key_press(self, widget, event):
        # ctrl+w means close window
        if (event.keyval == gdk.KEY_w \
            and event.state & gdk.ModifierType.CONTROL_MASK):
            self.close_window()

        # escape can mean several things
        if event.keyval == gdk.KEY_Escape:
            #check, maybe we are editing stuff
            if self.activityCell.get_property("editable"):
                self.activityCell.set_property("editable", False)
                return
            if self.categoryCell.get_property("editable"):
                self.categoryCell.set_property("editable", False)
                return

            self.close_window()

    """button events"""
    def on_category_add_clicked(self, button):
        """ appends row, jumps to it and allows user to input name """

        new_category = self.category_store.insert_before(self.category_store.unsorted_category,
                                                         [-2, _("New category")])

        model = self.category_tree.get_model()

        self.categoryCell.set_property("editable", True)
        self.category_tree.set_cursor_on_cell(model.get_path(new_category),
                                         focus_column = self.category_tree.get_column(0),
                                         focus_cell = None,
                                         start_editing = True)

    def on_activity_add_clicked(self, button):
        """ appends row, jumps to it and allows user to input name """
        category_id = self._get_selected_category()

        new_activity = self.activity_store.append([-1, _("New activity"), category_id])

        (model, iter) = self.selection.get_selected()

        self.activityCell.set_property("editable", True)
        self.activity_tree.set_cursor_on_cell(model.get_path(new_activity),
                                              focus_column = self.activity_tree.get_column(0),
                                              focus_cell = None,
                                              start_editing = True)

    def on_activity_remove_clicked(self, button):
        removable_id = self._del_selected_row(self.activity_tree)
        runtime.storage.remove_activity(removable_id)

    def on_day_start_changed(self, widget):
        day_start = self.day_start.time
        if day_start is None:
            return

        day_start = day_start.hour * 60 + day_start.minute
        conf.set("day-start-minutes", day_start)
    def on_close_button_clicked(self, button):
        self.close_window()



# pyTweener
#
# Tweening functions for python
#
# Heavily based on caurina Tweener: http://code.google.com/p/tweener/
#
# Released under M.I.T License - see above url
# Python version by Ben Harling 2009
# All kinds of slashing and dashing by Toms Baugis 2010, 2014
import math
import collections
import time
import re

from hamster.lib import datetime as dt

class Tweener(object):
    def __init__(self, default_duration = None, tween = None):
        """Tweener
        This class manages all active tweens, and provides a factory for
        creating and spawning tween motions."""
        self.current_tweens = collections.defaultdict(set)
        self.default_easing = tween or Easing.Cubic.ease_in_out
        self.default_duration = default_duration or 1.0

    def has_tweens(self):
        return len(self.current_tweens) > 0


    def add_tween(self, obj, duration = None, easing = None, on_complete = None,
                  on_update = None, round = False, delay = None, **kwargs):
        """
            Add tween for the object to go from current values to set ones.
            Example: add_tween(sprite, x = 500, y = 200, duration = 0.4)
            This will move the sprite to coordinates (500, 200) in 0.4 seconds.
            For parameter "easing" you can use one of the pytweener.Easing
            functions, or specify your own.
            The tweener can handle numbers, dates and color strings in hex ("#ffffff").
            This function performs overwrite style conflict solving - in case
            if a previous tween operates on same attributes, the attributes in
            question are removed from that tween.
        """
        if duration is None:
            duration = self.default_duration

        easing = easing or self.default_easing

        tw = Tween(obj, duration, delay, easing, on_complete, on_update, round, **kwargs )

        if obj in self.current_tweens:
            for current_tween in tuple(self.current_tweens[obj]):
                prev_keys = set((key for (key, tweenable) in current_tween.tweenables))
                dif = prev_keys & set(kwargs.keys())

                for key, tweenable in tuple(current_tween.tweenables):
                    if key in dif:
                        current_tween.tweenables.remove((key, tweenable))

                if not current_tween.tweenables:
                    current_tween.finish()
                    self.current_tweens[obj].remove(current_tween)


        self.current_tweens[obj].add(tw)
        return tw


    def get_tweens(self, obj):
        """Get a list of all tweens acting on the specified object
        Useful for manipulating tweens on the fly"""
        return self.current_tweens.get(obj, None)

    def kill_tweens(self, obj = None):
        """Stop tweening an object, without completing the motion or firing the
        on_complete"""
        if obj is not None:
            try:
                del self.current_tweens[obj]
            except:
                pass
        else:
            self.current_tweens = collections.defaultdict(set)

    def remove_tween(self, tween):
        """"remove given tween without completing the motion or firing the on_complete"""
        if tween.target in self.current_tweens and tween in self.current_tweens[tween.target]:
            self.current_tweens[tween.target].remove(tween)
            if not self.current_tweens[tween.target]:
                del self.current_tweens[tween.target]

    def finish(self):
        """jump the the last frame of all tweens"""
        for obj in self.current_tweens:
            for tween in self.current_tweens[obj]:
                tween.finish()
        self.current_tweens = {}

    def update(self, delta_seconds):
        """update tweeners. delta_seconds is time in seconds since last frame"""

        for obj in tuple(self.current_tweens):
            for tween in tuple(self.current_tweens[obj]):
                done = tween.update(delta_seconds)
                if done:
                    self.current_tweens[obj].remove(tween)
                    if tween.on_complete: tween.on_complete(tween.target)

            if not self.current_tweens[obj]:
                del self.current_tweens[obj]

        return self.current_tweens


class Tween(object):
    __slots__ = ('tweenables', 'target', 'delta', 'duration', 'delay',
                 'ease', 'delta', 'complete', 'round',
                 'on_complete', 'on_update')

    def __init__(self, obj, duration, delay, easing, on_complete, on_update, round,
                 **kwargs):
        """Tween object use Tweener.add_tween( ... ) to create"""

        #: should the tween values truncated to integers or not. Default is False.
        self.round = round

        #: duration of the tween
        self.duration = duration

        #: delay before the animation should be started
        self.delay = delay or 0

        self.target = obj

        #: easing function
        self.ease = easing

        # list of (property, start_value, delta)
        self.tweenables = set()
        for key, value in kwargs.items():
            self.tweenables.add((key, Tweenable(getattr(self.target, key), value)))

        self.delta = 0

        #: callback to execute on complete
        self.on_complete = on_complete

        #: callback to execute on update
        self.on_update = on_update

        self.complete = False

    def finish(self):
        self.update(self.duration)

    def update(self, ptime):
        """Update tween with the time since the last frame"""
        delta = self.delta + ptime
        total_duration = self.delay + self.duration

        if delta > total_duration:
            delta = total_duration

        if delta < self.delay:
            pass
        elif delta == total_duration:
            for key, tweenable in self.tweenables:
                setattr(self.target, key, tweenable.target_value)
        else:
            fraction = self.ease((delta - self.delay) / (total_duration - self.delay))

            for key, tweenable in self.tweenables:
                res = tweenable.update(fraction)
                if isinstance(res, float) and self.round:
                    res = int(res)
                setattr(self.target, key, res)

        if delta == total_duration or len(self.tweenables) == 0:
            self.complete = True

        self.delta = delta

        if self.on_update:
            self.on_update(self.target)

        return self.complete




class Tweenable(object):
    """a single attribute that has to be tweened from start to target"""
    __slots__ = ('start_value', 'change', 'decode_func', 'target_value', 'update')

    hex_color_normal = re.compile("#([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})")
    hex_color_short = re.compile("#([a-fA-F0-9])([a-fA-F0-9])([a-fA-F0-9])")


    def __init__(self, start_value, target_value):
        self.decode_func = lambda x: x
        self.target_value = target_value

        def float_update(fraction):
            return self.start_value + self.change * fraction

        def date_update(fraction):
            return dt.date.fromtimestamp(self.start_value + self.change * fraction)

        def datetime_update(fraction):
            return dt.datetime.fromtimestamp(self.start_value + self.change * fraction)

        def color_update(fraction):
            val = [max(min(self.start_value[i] + self.change[i] * fraction, 255), 0)  for i in range(3)]
            return "#%02x%02x%02x" % (val[0], val[1], val[2])


        if isinstance(start_value, int) or isinstance(start_value, float):
            self.start_value = start_value
            self.change = target_value - start_value
            self.update = float_update
        else:
            if isinstance(start_value, dt.datetime) or isinstance(start_value, dt.date):
                if isinstance(start_value, dt.datetime):
                    self.update = datetime_update
                else:
                    self.update = date_update

                self.decode_func = lambda x: time.mktime(x.timetuple())
                self.start_value = self.decode_func(start_value)
                self.change = self.decode_func(target_value) - self.start_value

            elif isinstance(start_value, str) \
             and (self.hex_color_normal.match(start_value) or self.hex_color_short.match(start_value)):
                self.update = color_update
                if self.hex_color_normal.match(start_value):
                    self.decode_func = lambda val: [int(match, 16)
                                                    for match in self.hex_color_normal.match(val).groups()]

                elif self.hex_color_short.match(start_value):
                    self.decode_func = lambda val: [int(match + match, 16)
                                                    for match in self.hex_color_short.match(val).groups()]

                if self.hex_color_normal.match(target_value):
                    target_value = [int(match, 16)
                                    for match in self.hex_color_normal.match(target_value).groups()]
                else:
                    target_value = [int(match + match, 16)
                                    for match in self.hex_color_short.match(target_value).groups()]

                self.start_value = self.decode_func(start_value)
                self.change = [target - start for start, target in zip(self.start_value, target_value)]



"""Robert Penner's classes stripped from the repetetive c,b,d mish-mash
(discovery of Patryk Zawadzki). This way we do the math once and apply to
all the tweenables instead of repeating it for each attribute
"""

def inverse(method):
    def real_inverse(t, *args, **kwargs):
        t = 1 - t
        return 1 - method(t, *args, **kwargs)
    return real_inverse

def symmetric(ease_in, ease_out):
    def real_symmetric(t, *args, **kwargs):
        if t < 0.5:
            return ease_in(t * 2, *args, **kwargs) / 2

        return ease_out((t - 0.5) * 2, *args, **kwargs) / 2 + 0.5
    return real_symmetric

class Symmetric(object):
    def __init__(self, ease_in = None, ease_out = None):
        self.ease_in = ease_in or inverse(ease_out)
        self.ease_out = ease_out or inverse(ease_in)
        self.ease_in_out = symmetric(self.ease_in, self.ease_out)


class Easing(object):
    """Class containing easing classes to use together with the tweener.
       All of the classes have :func:`ease_in`, :func:`ease_out` and
       :func:`ease_in_out` functions."""

    Linear = Symmetric(lambda t: t, lambda t: t)
    Quad = Symmetric(lambda t: t*t)
    Cubic = Symmetric(lambda t: t*t*t)
    Quart = Symmetric(lambda t: t*t*t*t)
    Quint = Symmetric(lambda t: t*t*t*t*t)
    Strong = Quint #oh i wonder why but the ported code is the same as in Quint

    Circ = Symmetric(lambda t: 1 - math.sqrt(1 - t * t))
    Sine = Symmetric(lambda t: 1 - math.cos(t * (math.pi / 2)))


    def _back_in(t, s=1.70158):
        return t * t * ((s + 1) * t - s)
    Back = Symmetric(_back_in)


    def _bounce_out(t):
        if t < 1 / 2.75:
            return 7.5625 * t * t
        elif t < 2 / 2.75:
            t = t - 1.5 / 2.75
            return 7.5625 * t * t + 0.75
        elif t < 2.5 / 2.75:
            t = t - 2.25 / 2.75
            return 7.5625 * t * t + .9375
        else:
            t = t - 2.625 / 2.75
            return 7.5625 * t * t + 0.984375
    Bounce = Symmetric(ease_out = _bounce_out)


    def _elastic_in(t, springiness = 0, wave_length = 0):
        if t in (0, 1):
            return t

        wave_length = wave_length or (1 - t) * 0.3

        if springiness <= 1:
            springiness = t
            s = wave_length / 4
        else:
            s = wave_length / (2 * math.pi) * math.asin(t / springiness)

        t = t - 1
        return -(springiness * math.pow(2, 10 * t) * math.sin((t * t - s) * (2 * math.pi) / wave_length))
    Elastic = Symmetric(_elastic_in)


    def _expo_in(t):
        if t in (0, 1): return t
        return math.pow(2, 10 * t) * 0.001
    Expo = Symmetric(_expo_in)



class _Dummy(object):
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

if __name__ == "__main__":
    import datetime as dt

    tweener = Tweener()
    objects = []

    object_count, update_times = 1000, 100

    for i in range(object_count):
        objects.append(_Dummy(i-100, i-100, i-100))


    total = dt.datetime.now()

    t = dt.datetime.now()
    print("Adding %d tweens..." % object_count)
    for i, o in enumerate(objects):
        tweener.add_tween(o, a = i,
                             b = i,
                             c = i,
                             duration = 0.1 * update_times,
                             easing=Easing.Circ.ease_in_out)
    print(dt.datetime.now() - t)

    t = dt.datetime.now()
    print("Updating %d times......" % update_times)
    for i in range(update_times):  #update 1000 times
        tweener.update(0.1)
    print(dt.datetime.now() - t)



# -*- coding: utf-8 -*-

# Copyright (C) 2008, 2014 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.

"""
License: GPLv2
"""

import logging
logger = logging.getLogger(__name__)   # noqa: E402

import os
from hamster.client import Storage

from gi.repository import Gdk as gdk
from gi.repository import Gio as gio
from gi.repository import GLib as glib
from gi.repository import GObject as gobject
from gi.repository import Gtk as gtk

import hamster

from hamster.lib import datetime as dt


class Controller(gobject.GObject):
    """Window creator and handler."""
    __gsignals__ = {
        "on-close": (gobject.SignalFlags.RUN_LAST, gobject.TYPE_NONE, ()),
    }

    def __init__(self, ui_file=""):
        gobject.GObject.__init__(self)

        if ui_file:
            self._gui = load_ui_file(ui_file)
            self.window = self.get_widget('window')
        else:
            self._gui = None
            self.window = gtk.Window()

        self.window.connect("delete-event", self.window_delete_event)
        if self._gui:
            self._gui.connect_signals(self)

    def get_widget(self, name):
        """ skip one variable (huh) """
        return self._gui.get_object(name)

    def window_delete_event(self, widget, event):
        self.close_window()

    def close_window(self):
        # Do not try to just hide;
        # dialogs are populated upon instanciation anyway
        self.window.destroy()
        self.window = None
        self.emit("on-close")

    def present(self):
        """Show window and bring it to the foreground."""
        # workaround https://gitlab.gnome.org/GNOME/gtk/issues/624
        # fixed in gtk-3.24.1 (2018-09-19)
        # self.overview_controller.window.present()
        self.window.present_with_time(glib.get_monotonic_time() / 1000)

    def show(self):
        """Show window.
        It might be obscurd by others though.
        See also: presents
        """
        self.window.show()

    def __bool__(self):
        return True if self.window else False


def load_ui_file(name):
    """loads interface from the glade file; sorts out the path business"""
    ui = gtk.Builder()
    ui.add_from_file(os.path.join(runtime.data_dir, name))
    return ui


class Singleton(object):
    def __new__(cls, *args, **kwargs):
        if '__instance' not in vars(cls):
            cls.__instance = object.__new__(cls, *args, **kwargs)
        return cls.__instance

class RuntimeStore(Singleton):
    """XXX - kill"""
    data_dir = ""
    home_data_dir = ""
    storage = None

    def __init__(self):
        self.version = hamster.__version__
        if hamster.installed:
            from hamster import defs  # only available when running installed
            self.data_dir = os.path.join(defs.DATA_DIR, "hamster")
        else:
            # running from sources
            module_dir = os.path.dirname(os.path.realpath(__file__))
            self.data_dir = os.path.join(module_dir, '..', '..', '..', 'data')

        self.data_dir = os.path.realpath(self.data_dir)
        self.storage = Storage()
        self.home_data_dir = os.path.realpath(os.path.join(glib.get_user_data_dir(), "hamster"))


runtime = RuntimeStore()


class GSettingsStore(gobject.GObject, Singleton):
    """
    Settings implementation which stores settings in GSettings
    Snatched from the conduit project (http://live.gnome.org/Conduit)
    """

    __gsignals__ = {
        "changed": (gobject.SignalFlags.RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT))
    }

    def __init__(self):
        gobject.GObject.__init__(self)
        self._settings = gio.Settings(schema_id='org.gnome.Hamster')

    def _key_changed(self, client, key, data=None):
        """
        Callback when a GSettings key changes
        """
        value = self._settings.get_value(key)
        self.emit('changed', key, value)

    def get(self, key, default=None):
        """
        Returns the value of the key or the default value if the key is
        not yet in GSettings
        """
        value = self._settings.get_value(key)
        if value is None:
            logger.warn("Unknown GSettings key: %s" % key)

        return value.unpack()

    def set(self, key, value):
        """
        Sets the key value in GSettings and connects adds a signal
        which is fired if the key changes
        """
        logger.debug("Settings %s -> %s" % (key, value))
        default = self._settings.get_default_value(key)
        assert default is not None
        self._settings.set_value(key, glib.Variant(default.get_type().dup_string(), value))
        return True

    def bind(self, key, obj, prop):
        self._settings.bind(key, obj, prop, gio.SettingsBindFlags.DEFAULT)

    @property
    def day_start(self):
        """Start of the hamster day."""
        day_start_minutes = self.get("day-start-minutes")
        hours, minutes = divmod(day_start_minutes, 60)
        return dt.time(hours, minutes)


conf = GSettingsStore()



# -*- coding: utf-8 -*-

# Copyright (C) 2007, 2008 Toms Bauģis <toms.baugis at gmail.com>

# This file is part of Project Hamster.

# Project Hamster is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Project Hamster is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Project Hamster.  If not, see <http://www.gnu.org/licenses/>.


from os.path import join
from hamster.lib.configuration import runtime
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk

class About(object):
    def __init__(self, parent=None):
        about = gtk.AboutDialog(parent=parent)
        self.window = about
        infos = {
            "program-name" : "Hamster",
            "version" : runtime.version,
            "comments" : _("Project Hamster — track your time"),
            "copyright" : _("Copyright © 2007–2010 Toms Bauģis and others"),
            "website" : "https://github.com/projecthamster/hamster/wiki/",
            "website-label" : _("Project Hamster Website"),
            "title": _("About Time Tracker"),
            "wrap-license": True
        }

        about.set_authors(["Toms Bauģis <toms.baugis@gmail.com>",
                           "Patryk Zawadzki <patrys@pld-linux.org>",
                           "Pēteris Caune <cuu508@gmail.com>",
                           "Juanje Ojeda <jojeda@emergya.es>"])
        about.set_artists(["Kalle Persson <kalle@kallepersson.se>"])

        about.set_translator_credits(_("translator-credits"))

        for prop, val in infos.items():
            about.set_property(prop, val)

        about.set_logo_icon_name("org.gnome.Hamster.GUI")

        about.connect("response", lambda self, *args: self.destroy())
        about.show_all()

    def present(self):
        self.window.present()
